# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import math

from mathmakerlib.exceptions import ZERO_OBJECTS_ERRORS
from mathmakerlib.geometry.point import Point
from mathmakerlib.calculus.number import Number
from mathmakerlib.core.dimensional import Dimensional
from mathmakerlib.calculus.tools import is_number


class Vector(Dimensional):
    """
    A euclidean (free) vector.

    Differences with Bipoints are discussed in Bipoint docstring.

    This class won't ever need to get Drawable, but can be instanciated.
    """

    def __init__(self, *args, allow_zero_length=True):
        """
        It's possible to create a Vector giving:
        - a Bipoint: Bipoint(A, B)
        - a pair of Points: A, B
        - its coordinates x, y or x, y, z
        """
        if not args or len(args) >= 4:
            raise TypeError('Vector() takes one, two or three arguments '
                            '({} given)'.format(len(args)))
        if len(args) == 1:
            from mathmakerlib.geometry.bipoint import Bipoint
            if not isinstance(args[0], Bipoint):
                raise TypeError('a Vector can be created from one Bipoint, '
                                'found {} instead.'.format(repr(args[0])))
            self._x = args[0].Δx
            self._y = args[0].Δy
            if args[0].three_dimensional:
                self._z = args[0].Δz
                self._three_dimensional = True
            else:
                self._z = Number(0)
                self._three_dimensional = False
        elif len(args) == 2:
            # Two Points
            if isinstance(args[0], Point) and isinstance(args[1], Point):
                self._x = args[1].x - args[0].x
                self._y = args[1].y - args[0].y
                if args[0].three_dimensional or args[1].three_dimensional:
                    self._three_dimensional = True
                    self._z = args[1].z - args[0].z
                else:
                    self._three_dimensional = False
                    self._z = Number(0)
            # Two numbers
            elif is_number(args[0]) and is_number(args[1]):
                self._three_dimensional = False
                self._x = Number(args[0])
                self._y = Number(args[1])
                self._z = Number(0)
            else:
                raise TypeError('a Vector can be created from two arguments, '
                                'either two Points or two numbers. '
                                'Found {} and {} instead.'
                                .format(repr(args[0]), repr(args[1])))
        elif len(args) == 3:
            self._three_dimensional = True
            self._x = Number(args[0])
            self._y = Number(args[1])
            self._z = Number(args[2])
        self._length = Number(self.x ** 2 + self.y ** 2 + self.z ** 2).sqrt()
        if not allow_zero_length and self.length == 0:
            msg = 'Explicitly disallowed creation of a zero-length {}.'\
                .format(type(self).__name__)
            raise ZERO_OBJECTS_ERRORS[type(self).__name__](msg)

    def __repr__(self):
        if self.three_dimensional:
            return 'Vector({}, {}, {})'.format(str(self.x), str(self.y),
                                               str(self.z))
        else:
            return 'Vector({}, {})'.format(str(self.x), str(self.y))

    def __eq__(self, other):
        if isinstance(other, Vector):
            return (self.x == other.x and self.y == other.y
                    and self.z == other.z)
        else:
            return False

    def __add__(self, other):
        if not isinstance(other, Vector):
            raise TypeError('Can only add a Vector to another Vector. '
                            'Found {} instead.'.format(repr(other)))
        if self.three_dimensional or other.three_dimensional:
            return Vector(self.x + other.x,
                          self.y + other.y,
                          self.z + other.z)
        else:
            return Vector(self.x + other.x,
                          self.y + other.y)

    def __neg__(self):
        if self.three_dimensional:
            return Vector(-self.x, -self.y, -self.z)
        else:
            return Vector(-self.x, -self.y)

    def dot(self, other):
        if not isinstance(other, Vector):
            raise TypeError('Can only calculate the dot product of a '
                            'Vector by another Vector. Found {} instead.'
                            .format(repr(other)))
        return self.x * other.x + self.y * other.y + self.z * other.z

    def cross(self, other):
        if not isinstance(other, Vector):
            raise TypeError('Can only calculate the cross product of a '
                            'Vector by another Vector. Found {} instead.'
                            .format(repr(other)))
        return Vector(self.y * other.z - self.z * other.y,
                      self.z * other.x - self.x * other.z,
                      self.x * other.y - self.y * other.x)

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    @property
    def z(self):
        return self._z

    @property
    def coordinates(self):
        return (self._x, self._y, self._z)

    @property
    def length(self):
        """Vector's norm."""
        return self._length

    def normalized(self):
        """Return the unit Vector colinear (to self)."""
        if self.three_dimensional:
            return Vector(self.x / self.length,
                          self.y / self.length,
                          self.z / self.length)
        else:
            return Vector(self.x / self.length,
                          self.y / self.length)

    def _slope(self, offset=0):
        if self.length == 0:
            msg = 'Cannot calculate the slope of a zero-length {}.'\
                .format(type(self).__name__)
            raise ZERO_OBJECTS_ERRORS[type(self).__name__](msg)
        theta = Number(
            str(math.degrees(math.acos(self.x / self.length))))\
            .rounded(Number('0.001'))
        return theta if self.y >= 0 else Number(offset) - theta

    @property
    def slope(self):
        """Slope between the Vector and X-axis, from -180° to 180°."""
        return self._slope()

    @property
    def slope360(self):
        """Slope between the Vector and X-axis, from 0° to 360°."""
        return self._slope(360)

    def angle_measure(self, other):
        """Angle between the Vector and another Vector."""
        result = other.slope360 - self.slope360
        if result < 0:
            result += 360
        return result

    def bisector(self, other, new_endpoint_name='automatic'):
        """
        Return the bisector vector of self and another vector.

        :param arg: the other Vector
        :type arg: Vector
        """
        if not isinstance(other, Vector):
            raise TypeError('Can only create the bisector with another '
                            'Vector. Found {} instead.'
                            .format(repr(other)))
        result = self.normalized() + other.normalized()
        if self.angle_measure(other) > 180:
            result = -result
        return result



# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from .tools import convex_hull
from .point import Point
from .bipoint import Bipoint
from .vector import Vector
from .linesegment import LineSegment
from .dividedlinesegment import DividedLineSegment
from .xaxis import XAxis
from .polygons import Polygon, shoelace_formula
from .polygons import Triangle, RightTriangle, EquilateralTriangle
from .polygons import IsoscelesTriangle
from .polygons import Quadrilateral, Rectangle, Rhombus, Square
from .angle import AngleDecoration, Angle, AnglesSet
from .polyhedra import Polyhedron, RightCuboid
from .projections import ObliqueProjection

__all__ = ['convex_hull',
           'Point', 'Bipoint', 'Vector', 'LineSegment', 'DividedLineSegment',
           'XAxis',
           'Polygon', 'shoelace_formula',
           'Triangle', 'RightTriangle', 'EquilateralTriangle',
           'IsoscelesTriangle',
           'Quadrilateral', 'Rectangle', 'Rhombus', 'Square',
           'AngleDecoration', 'Angle', 'AnglesSet',
           'Polyhedron', 'RightCuboid',
           'ObliqueProjection']



# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


def convex_hull(*points):
    """
    Compute the convex hull of a set of points belonging to plane Oxy.

    Mostly taken from:
    https://en.wikibooks.org/wiki/Algorithm_Implementation/Geometry
    /Convex_hull/Monotone_chain

    Output: a list of vertices of the convex hull in counter-clockwise order,
    starting from the vertex with the smallest coordinates.

    Implements Andrew's monotone chain algorithm. O(n log n) complexity.

    :param points: a list of Points.
    :type points: a list of Points.
    """
    # Remove duplicates to detect the case we have just one unique point.
    # Sort the points by abscissa, then ordinate, then applicate.
    points = sorted(set(points), key=lambda point: point.coordinates)

    # Boring case: no points or a single point, possibly repeated multiple
    # times.
    if len(points) <= 1:
        return points

    # 2D cross product of OA and OB vectors, i.e. z-component of their 3D
    # cross product.
    # Returns a positive value, if OAB makes a counter-clockwise turn,
    # negative for clockwise turn, and zero if the points are collinear.
    def cross(o, a, b):
        return (a.x - o.x) * (b.y - o.y) - (a.y - o.y) * (b.x - o.x)

    # Build lower hull
    lower = []
    for p in points:
        while len(lower) >= 2 and cross(lower[-2], lower[-1], p) <= 0:
            lower.pop()
        lower.append(p)

    # Build upper hull
    upper = []
    for p in reversed(points):
        while len(upper) >= 2 and cross(upper[-2], upper[-1], p) <= 0:
            upper.pop()
        upper.append(p)

    # Concatenation of the lower and upper hulls gives the convex hull.
    # Last point of each list is omitted because it is repeated at the
    # beginning of the other list.
    return lower[:-1] + upper[:-1]



# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import copy
import math
import locale
import random
import warnings
from decimal import Decimal, ROUND_DOWN, ROUND_HALF_UP

from mathmakerlib import required
from mathmakerlib.calculus.unit import physical_quantity
from mathmakerlib.calculus.tools import is_number, is_integer
from mathmakerlib.core.signed import Signed
from mathmakerlib.core.printable import Printable
from mathmakerlib.core.evaluable import Evaluable
from mathmakerlib.calculus.unit import Unit, difference_of_orders_of_magnitude


class Sign(Printable, Evaluable):

    def __init__(self, o):
        self._sign = '+'
        self.sign = o

    def __repr__(self):
        return 'Sign({})'.format(self.sign)

    def __str__(self):
        return str(self.sign)

    def __eq__(self, other):
        if isinstance(other, Sign):
            return self.sign == other.sign
        elif other in ['+', '-']:
            return self.sign == other
        else:
            return False

    def __ne__(self, other):
        if isinstance(other, Sign):
            return self.sign != other.sign
        elif other in ['+', '-']:
            return self.sign != other
        else:
            return True

    def imprint(self, start_expr=True, variant='latex'):
        return self.sign

    @property
    def sign(self):
        return self._sign

    @sign.setter
    def sign(self, o):
        if o in ['+', '-']:
            self._sign = o
        elif isinstance(o, Sign):
            self._sign = o._sign
        elif isinstance(o, Number):
            if o >= 0:
                self._sign = '+'
            else:
                self._sign = '-'
        else:
            raise ValueError('o must be \'+\', \'-\' or a Number.')

    def __mul__(self, o):
        if isinstance(o, Sign):
            if self.sign == o.sign:
                return Sign('+')
            else:
                return Sign('-')
        elif isinstance(o, Number):
            if self.sign == '+':
                return o
            else:
                return -o
        else:
            raise TypeError('Cannot multiply a Sign by a {}.'
                            .format(str(type(o))))

    def __rmul__(self, o):
        return self.__mul__(o)

    def evaluate(self, **kwargs):
        if self.sign == '+':
            return Number(1)
        else:
            return Number(-1)


class Number(Decimal, Signed, Printable, Evaluable):
    """Extend Decimal with a bunch of useful methods."""

    # To keep immutability of Decimal, use __new__ not __init__
    def __new__(cls, value='0', context=None, unit='undefined'):
        if isinstance(value, Sign):
            value = value.evaluate()
        if isinstance(value, float):
            value = str(value)
        if isinstance(value, Number) and unit == 'undefined':
            unit = copy.deepcopy(value.unit)
        if unit == 'undefined':
            unit = None
        self = super().__new__(cls, value=value, context=context)
        if unit is None:
            self._unit = None
        else:
            self._unit = Unit(unit)
        return self

    def __hash__(self):
        return hash(Decimal(self))

    def __copy__(self):  # Similar to Decimal.__copy__()
        if type(self) is Number:
            return self     # I'm immutable; therefore I am my own clone
        return self.__class__(self)

    def __deepcopy__(self, memo):  # Similar to Decimal.__deepcopy__()...
        if type(self) is Number:
            # ...but my components are *not* also immutable
            return Number(value=self, unit=self.unit)
        return self.__class__(self, unit=self.unit)

    def __eq__(self, other):
        if self.unit is None:
            return Decimal(self) == other
        else:
            if not isinstance(other, Number):
                return False
            else:
                return (Decimal(self) == Decimal(other)
                        and self.unit == other.unit)

    def __ne__(self, other):
        if self.unit is None:
            return Decimal(self) != other
        else:
            if not isinstance(other, Number):
                return True
            else:
                return not (Decimal(self) == Decimal(other)
                            and self.unit == other.unit)

    def __add__(self, other, context=None):
        if isinstance(other, Sign):
            raise TypeError('Cannot add a Sign and a Number')
        if not isinstance(other, Number):
            other = Number(other)
        if self.unit == other.unit:
            return Number(Decimal(self).__add__(other), unit=self.unit)
        else:
            if hasattr(other, 'unit') and other.unit is not None:
                other_unit = str(other.unit)
            else:
                other_unit = str(None)
            raise ValueError('Cannot add two Numbers having different '
                             'Units ({} and {}).'
                             .format(str(self.unit),
                                     other_unit))

    def __radd__(self, other, context=None):
        return self.__add__(other, context=context)

    def __sub__(self, other, context=None):
        if isinstance(other, Sign):
            raise TypeError('Cannot subtract a Sign and a Number')
        if not isinstance(other, Number):
            other = Number(other)
        if self.unit == other.unit:
            return Number(Decimal(self).__sub__(other), unit=self.unit)
        else:
            if hasattr(other, 'unit') and other.unit is not None:
                other_unit = str(other.unit)
            else:
                other_unit = str(None)
            raise ValueError('Cannot subtract two Numbers having different '
                             'Units ({} and {}).'
                             .format(str(self.unit),
                                     other_unit))

    def __rsub__(self, other, context=None):
        return -self.__sub__(other, context=context)

    def __mul__(self, other, context=None):
        if isinstance(other, Sign):
            other = other.evaluate()
        other = Number(other)
        u = None
        if self.unit is None:
            u = copy.deepcopy(other.unit)
        elif other.unit is None:
            u = copy.deepcopy(self.unit)
        elif self.unit.content == other.unit.content:
            if self.unit.exponent is None:
                se = Number(1)
            else:
                se = Number(self.unit.exponent)
            if other.unit.exponent is None:
                oe = Number(1)
            else:
                oe = Number(other.unit.exponent)
            u = Unit(self.unit.content, exponent=se + oe)
        else:
            if self.unit.exponent is None or self.unit.exponent == Number(1):
                u = Unit(self.unit.content + '.' + other.unit.content,
                         exponent=other.unit.exponent)
            elif (other.unit.exponent is None
                  or other.unit.exponent == Number(1)):
                u = Unit(other.unit.content + '.' + self.unit.content,
                         exponent=self.unit.exponent)
            else:
                raise NotImplementedError('Cannot yet handle a '
                                          'multiplication of {} by {}.'
                                          .format(repr(self), repr(other)))
        return Number(Decimal(self).__mul__(other),
                      unit=u)

    def __rmul__(self, other, context=None):
        return self.__mul__(other, context=context)

    def __truediv__(self, other, context=None):
        if isinstance(other, Sign):
            other = other.evaluate()
        other = Number(other)
        u = None
        if self.unit is None and other.unit is None:
            u = None
        elif self.unit is None:
            u = copy.deepcopy(other.unit)
            if u.exponent is None:
                u.exponent = -Number(1)
            else:
                u.exponent = -u.exponent
        elif other.unit is None:
            u = copy.deepcopy(self.unit)
        elif self.unit.content == other.unit.content:
            if self.unit.exponent is None:
                se = Number(1)
            else:
                se = Number(self.unit.exponent)
            if other.unit.exponent is None:
                oe = Number(1)
            else:
                oe = Number(other.unit.exponent)
            if se == oe:
                u = None
            elif se - oe == Number(1):
                u = Unit(self.unit.content, exponent=None)
            else:
                u = Unit(self.unit.content, exponent=se - oe)
        else:
            if ((self.unit.exponent is None or self.unit.exponent == Number(1))
                and (other.unit.exponent is None
                     or other.unit.exponent == Number(1))):
                u = Unit(self.unit.content + '/' + other.unit.content)
            else:
                raise NotImplementedError('Cannot yet handle a '
                                          'division of {} by {}.'
                                          .format(repr(self), repr(other)))
        return Number(Decimal(self).__truediv__(other), unit=u)

    def __rtruediv__(self, other, context=None):
        return Number.__truediv__(Number(other), self, context=context)

    def __floordiv__(self, other, context=None):
        if isinstance(other, Sign):
            other = other.evaluate()
        return Number(Decimal(self).__floordiv__(other))

    def __rfloordiv__(self, other, context=None):
        return Number.__floordiv__(Number(other), self, context=context)

    def __mod__(self, other, context=None):
        return Number(Decimal(self).__mod__(other))

    def __rmod__(self, other, context=None):
        return Number(Decimal(self).__rmod__(other))

    def __divmod__(self, other, context=None):
        r = Decimal(self).__divmod__(other)
        return (Number(r[0]), Number(r[1]))

    def __rdivmod__(self, other, context=None):
        r = Decimal(self).__rdivmod__(other)
        return (Number(r[0]), Number(r[1]))

    def __pow__(self, other, context=None):
        return Number(Decimal(self).__pow__(other))

    def __rpow__(self, other, context=None):
        return Number(Decimal(self).__rpow__(other))

    def __neg__(self):
        return Number(-Decimal(self), unit=copy.deepcopy(self.unit))

    def __pos__(self):
        return Number(+Decimal(self), unit=copy.deepcopy(self.unit))

    def __abs__(self):
        return Number(abs(Decimal(self)), unit=copy.deepcopy(self.unit))

    def __str__(self):
        basic_str = Decimal.__str__(self)
        if self.unit is None:
            return basic_str
        else:
            return basic_str + ' ' + str(self.unit)

    def __repr__(self):
        basic_repr = repr(Decimal(Decimal.__str__(self))).replace('Decimal',
                                                                  'Number')
        if self.unit is None:
            return basic_repr
        else:
            return basic_repr.replace("')", " {}')"
                                            .format(self.unit.uiprinted))

    def sqrt(self):
        return Number(Decimal(self).sqrt())

    def quantize(self, exp, rounding=None, context=None):
        return Number(Decimal(self).quantize(exp,
                                             rounding=rounding,
                                             context=context),
                      context=context,
                      unit=self.unit)

    @property
    def unit(self):
        return self._unit

    @property
    def sign(self):
        if self < 0:
            return Sign('-')
        else:
            return Sign('+')

    def imprint(self, start_expr=True, variant='latex', dot=False):
        extra_sign = ''
        if not start_expr and self >= 0:
            extra_sign = '+'

        if variant in ['latex', 'siunitx']:
            self_str = locale.format_string(
                '%.{}f'.format(
                    self.fracdigits_nb(ignore_trailing_zeros=False)),
                Decimal(self))
        elif variant == 'user_input':
            self_str = Decimal.__str__(self)
        else:
            raise ValueError('variant must belong to [\'latex\', \'siunitx\', '
                             '\'user_input\']; got \'{}\' instead.'
                             .format(variant))
        if dot:
            self_str = self_str\
                .replace(locale.localeconv()['decimal_point'], '.')
        if self.unit is None:
            if variant == 'siunitx':
                required.package['siunitx'] = True
                return r'\num{{{}}}'.format(extra_sign + self_str)
            else:
                return extra_sign + self_str
        else:
            if variant == 'latex':
                required.package['siunitx'] = True
                if self.unit.content in ['€', r'\officialeuro']:
                    required.package['eurosym'] = True
                if physical_quantity(self.unit) == 'angle':
                    return extra_sign + r'\ang{' + self_str + '}'
                else:
                    return extra_sign + r'\SI{' + self_str \
                        + '}{' + self.unit.imprint(standalone=False) + '}'
            else:  # 'user_input'
                return extra_sign + self_str + ' ' + self.unit.uiprinted

    def evaluate(self, **kwargs):
        return self

    def standardized(self):
        """Turn 8.0 to 8 and 1E+1 to 10"""
        return Number(self.quantize(Decimal(1)), unit=self.unit) \
            if self == self.to_integral() \
            else Number(self.normalize(), unit=self.unit)

    def converted_to(self, unit):
        if isinstance(unit, str):
            unit = Unit(unit)
        try:
            factor = difference_of_orders_of_magnitude(self.unit, unit)
        except TypeError as excinfo:
            if str(excinfo).startswith('Cannot give the difference of orders '
                                       'of magnitude between two units that '
                                       'do not belong to the same '
                                       'physical quantity'):
                raise TypeError('Cannot convert {} into {}.'
                                .format(self.uiprinted, unit.uiprinted))
        return Number(self * factor, unit=unit)

    def lowest_nonzero_digit_index(self):
        """
        Index of the lowest digit different from zero.

        Indices are 0 for the unit position, 1 for tenths, 2 for hundredths,
        etc. and -1 for tens, -2 for hundreds etc.

        If self is 0, then return None.
        """
        if self == 0:
            return None
        fdn = self.fracdigits_nb()
        if fdn > 0:
            return fdn
        n_tenth = self / 10
        if n_tenth.fracdigits_nb() == 0:
            return -1 + n_tenth.lowest_nonzero_digit_index()
        else:
            return 0

    def nonzero_digits_nb(self):
        """Return the number of nonzero digits."""
        n = str(abs(self.standardized()))
        return len(n) - n.count('0') - n.count('.')

    def isolated_zeros(self):
        """Return the number of zeros inside other digits."""
        if self == 0:
            return 0
        elif self % 10 == 0:
            return Number(self // 10).isolated_zeros()
        else:
            return self.standardized().as_tuple().digits.count(0)

    def rounded(self, precision, rounding=ROUND_HALF_UP):
        """
        Round the number. Return a standardized result.

        :param precision: a Decimal (same as decimal.Decimal().quantize()).
                          For instance, Decimal('1'), Decimal('1.0') etc.
        """
        if precision >= 10:
            rounded_val = round(self, -int(math.log10(precision)))
        else:
            rounded_val = self.quantize(precision, rounding=rounding)
        return Number(rounded_val, unit=self.unit).standardized()

    def fracdigits_nb(self, ignore_trailing_zeros=True):
        """Return the number of fractional digits."""
        n = Number(Decimal(self), unit=None)
        if ignore_trailing_zeros:
            n = Number(abs(n)).standardized()
        else:
            n = abs(n)
        temp = len(str((n - n.rounded(Decimal(1), rounding=ROUND_DOWN)))) - 2
        return temp if temp >= 0 else 0

    def digits_sum(self):
        """Return the sum of all digits."""
        _, digits, e = self.standardized().as_tuple()
        return sum(list(digits))

    def is_power_of_10(self):
        """Check if n is a power of ten."""
        n = Number(abs(self))
        if Decimal(10) <= n:
            return Number(n / Number(10)).is_power_of_10()
        if Decimal(1) < n < Decimal(10):
            return False
        elif n == Decimal(1):
            return True
        elif Decimal('0.1') < n < Decimal(1):
            return False
        elif 0 < n <= Decimal('0.1'):
            return Number(n * Number(10)).is_power_of_10()
        elif n == 0:
            return False

    def is_pure_half(self):
        """True if fractional part is .5"""
        return self % 1 == Number('0.5')

    def is_pure_quarter(self):
        """True if fractional part is .25 or .75"""
        return self % 1 in (Number('0.25'), Number('0.75'))

    @property
    def digits(self):
        """Break down the number into a dictionary."""
        _, digits, e = self.standardized().as_tuple()
        return {Number(10) ** (e + len(digits) - 1 - i): d
                for i, d in enumerate(digits)}

    def highest_digitplace(self):
        """Return a power of ten matching the highest digitplace."""
        _, digits, e = self.standardized().as_tuple()
        return Number(10) ** (e + len(digits) - 1)

    def estimation(self):
        """Round to the highest digitplace possible."""
        return self.rounded(self.highest_digitplace())

    def digit(self, digitplace):
        """
        Return the digit matching digitplace.

        :param digitplace: a power of ten matching the digit to get
        :type digitplace: Decimal
        """
        if not Number(digitplace).is_power_of_10():
            raise ValueError('Expect a power of ten, found {} instead.'
                             .format(repr(digitplace)))
        if digitplace in self.digits:
            return self.digits[digitplace]
        else:
            return 0

    def atomized(self, keep_zeros=False):
        """Split abs(self) in as many Numbers as digits."""
        _, digits, e = self.standardized().as_tuple()
        digits = list(digits)
        result = []
        for i, d in enumerate(digits):
            if d != 0 or keep_zeros:
                result += [Number(d) * Number(10) ** (e + len(digits) - 1 - i)]
        if not len(result):
            return [Number(0)]
        return [Number(n) for n in result]

    def overlap_level(self):
        """
        Calculate the maximum overlap possible.

        For instance, 0.724 has an overlap level of 1 (can be split as
        0.71 + 0.14) but 0.714 has an overlap level of only 0 (can be split as
        0.7 + 0.014 or 0.71 + 0.004, but not as two decimals having a nonzero
        digits at tenths' position.)
        """
        base_level = self.standardized().nonzero_digits_nb() - 2
        digits = list(self.as_tuple().digits)
        level = base_level - digits[1:-1].count(1)
        return level

    def cut(self, overlap=0, return_all=False):
        """
        Cut self as a + b, a and b having only 0, 1... positions in common.

        For instance:
        Number('5.6').cut() == (Number('5'), Number('0.6'))
        Number('5.36').cut(return_all=True) == \
            [(Number('5'), Number('0.36')),
             (Number('5.3'), Number('0.06'))]
        Number('5.36').cut(overlap=1, return_all=True) == \
            [(Number('5.1'), Number('0.26')),
             (Number('5.2'), Number('0.16'))]

        :param overlap: tells how many decimal places in common the terms a
                        and b should have when splitting as a sum.
                        Values of overlap >= 2 are not handled yet
        :type overlap: int (only 0 or 1 are handled yet)
        :param return_all: if True, then all possibilities are returned, as a
                           list.
        :type return_all: bool
        :rtype: tuple (or a list of tuples)
        """
        if not isinstance(overlap, int):
            raise TypeError('When overlap is used, it must be an int. Got '
                            'a {} instead.'.format(type(overlap)))
        if overlap < 0:
            raise ValueError('overlap must be a positive int. Got a '
                             'negative int instead.')
        if self.overlap_level() < overlap:
            raise ValueError('Given overlap is too high.')
        if overlap not in [0, 1]:
            raise ValueError('Only 0 <= overlap <= 1 is implemented yet.')
        results = []
        digits = self.atomized()
        if overlap == 0:
            for n in range(len(digits) - 1):
                results += [(sum(digits[0:n + 1]), sum(digits[n + 1:]))]
        elif overlap == 1:
            for i in range(self.overlap_level()):
                results += [(sum(digits[0:i + 1] + [d[0]]),
                             sum(digits[i + 2:]) + d[1])
                            for d
                            in digits[i + 1].split(return_all=True)]
        if return_all:
            return results
        else:
            return random.choice(results)

    def split(self, operation='sum', dig=0, return_all=False,
              int_as_halves=False, int_as_quarters=False,
              int_as_halves_or_quarters=False, integer_split_at_unit=False):
        """
        Split self as a sum or difference, e.g. self = a + b or self = a - b

        By default, a and b have as many fractional digits as n does.
        The 'dig' keyword tells how many extra digits must have a and b
        (compared to self). For instance, if n=Decimal('4.5'), operation='sum',
        dig=1, then self will be split into 2-fractional-digits numbers,
        like 2.14 + 2.36.

        :param operation: must be 'sum', 'difference', '+' or '-'
        :type operation: str
        :param dig: extra depth level to use
        :type dig: int
        :param int_as_halves: whether integers should be split as two integers
        ± 0.5. Disabled if self is not integer.
        :type int_as_halves: bool
        :param int_as_quarters: whether integers should be split as two
        integers ± 0.25 Disabled if self is not integer.
        :type int_as_quarters: bool
        :param int_as_halves_or_quarters: whether integers should be split as
        two integers ± 0.5 or ±0.25 (randomly). Disabled if self is not
        integer.
        :type int_as_halves_or_quarters: bool
        :rtype: tuple (of numbers)
        """
        if operation not in ['sum', 'difference', '+', '-']:
            raise ValueError('Argument "operation" should be either \'sum\' '
                             'or \'difference\'.')
        if integer_split_at_unit:
            n_depth = self.fracdigits_nb()
            depth = dig + self.fracdigits_nb()
        else:
            n_depth = self.lowest_nonzero_digit_index()
            depth = dig + self.lowest_nonzero_digit_index()
        n = self
        if operation in ['sum', '+']:
            if self.is_power_of_10() and abs(self) <= 1 and dig == 0:
                # This case is impossible: write 1 as a sum of two natural
                # numbers bigger than 1, or 0.1 as a sum of two positive
                # decimals having 1 digit either, etc. so we arbitrarily
                # replace self by a random number between 2 and 9
                warnings.warn('mathmakerlib is asked something impossible '
                              '(split {} as a sum of two numbers having as '
                              'many digits)'.format(self))
                n = random.choice([i + 2 for i in range(7)])
                n = n * (10 ** (Decimal(- n_depth)))
            amplitude = n
        elif operation in ['difference', '-']:
            amplitude = max(10 ** (n_depth), n)
        start, end = 0, int((amplitude) * 10 ** depth - 1)
        if start > end:
            start, end = end + 1, -1
        # default: all numbers, including integers
        seq = [(Decimal(i) + 1) / Decimal(10) ** Decimal(depth)
               for i in range(start, end)]
        # then if decimals are wanted, we remove the results that do not match
        # the wanted "depth" (if depth == 2, we remove 0.4 for instance)
        if depth >= 1:
            seq = [i for i in seq
                   if not is_integer(i * (10 ** (depth - 1)))]
        delta = 0
        if is_integer(self):
            if int_as_halves_or_quarters:
                h, q = (True, False) if random.choice([True, False]) \
                    else (False, True)
                int_as_halves, int_as_quarters = h, q
            if int_as_halves:
                delta = Number('0.5')
            if int_as_quarters:
                delta = Number('0.25')
        if return_all:
            if operation in ['sum', '+']:
                return [(Number(a) + delta, Number(n - a) - delta)
                        for a in seq]
            elif operation in ['difference', '-']:
                return [(Number(n + b) + delta, Number(b) + delta)
                        for b in seq]
        else:
            if operation in ['sum', '+']:
                a = random.choice(seq)
                b = n - a
                a += delta
                b -= delta
            elif operation in ['difference', '-']:
                b = random.choice(seq)
                a = n + b
                a += delta
                b += delta
            return (Number(a), Number(b))


def move_fracdigits_to(n, from_nb=None):
    """
    Turn n into decimal instead of all decimals found in the from_nb list.

    Each decimal found in the numbers' list will be recursively replaced by
    10 times itself (until it is no decimal anymore) while in the same time
    n will be divided by 10.

    This is useful for the case division by a decimal is unwanted.

    :param n: the number who will be divided by 10 instead of the others
    :type n: any number (int, Decimal, float though they're not advised)
    :param from_nb: an iterable containing the numbers that must be integers
    :type from_nb: a list (of numbers)
    :rtype: a list (of numbers)
    """
    if type(from_nb) is not list:
        raise TypeError('A list of numbers must be given as argument '
                        '\'numbers\'.')
    if not is_number(n):
        raise TypeError('The first argument must be a number.')
    n = Decimal(str(n))
    # If any element of from_nb is not a number, is_integer() will raise
    # an exception.
    if all([is_integer(i) for i in from_nb]):
        return [n, ] + [i for i in from_nb]
    numbers_copy = copy.deepcopy(from_nb)
    for i, j in enumerate(from_nb):
        if not is_integer(j):
            numbers_copy[i] = j * 10
            return move_fracdigits_to(n / 10, from_nb=numbers_copy)


def remove_fracdigits_from(number, to=None):
    """
    Turn a number of the to list into a decimal, instead of number.

    In some cases this is not possible (for instance when all numbers of the
    to list are multiples of 10), then a ValueError is raised.

    :param number: the number that must be turned into integer
    :type number: decimal.Decimal
    :param to: the list of numbers where to find an integer that must be
               turned into a decimal
    :type to: list
    :rtype: a list (of numbers)
    """
    if not isinstance(number, Decimal):
        raise TypeError('The first argument must be a Decimal number.')
    if is_integer(number):
        raise TypeError('The first argument must be a decimal number.')
    if type(to) is not list:
        raise TypeError('Argument to: must be a list.')
    n = Number(number).fracdigits_nb()
    try:
        i = to.index(next(x for x in to
                          if not is_integer(x / 10 ** n)))
    except StopIteration:
        raise ValueError('None of the numbers of to can be turned into '
                         'decimal.')
    to[i] = Number(to[i]) / Number(10) ** n
    return [Number(number * 10 ** n).standardized(), ] + [x for x in to]


def fix_fracdigits(n1, *n2):
    """Ensure digits from n1 are removed. Change n2 if necessary."""
    n2 = list(n2)
    try:
        n1, *n2 = remove_fracdigits_from(n1, to=n2)
    except ValueError:
        j = random.choice([j for j in range(len(n2))])
        n2[j] += random.choice([i for i in range(-4, 5) if i != 0])
        n1, *n2 = remove_fracdigits_from(n1, to=n2)
    return (n1, *n2)



# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from .tools import is_number, is_integer, is_natural, prime_factors
from .exponented import Exponented
from .number import Sign, Number, move_fracdigits_to, remove_fracdigits_from
from .number import fix_fracdigits
from .unit import Unit, physical_quantity, difference_of_orders_of_magnitude
from .fraction import Fraction
from .clocktime import ClockTime

__all__ = ['is_number', 'is_integer', 'is_natural', 'Exponented', 'Sign',
           'Number', 'move_fracdigits_to', 'remove_fracdigits_from',
           'fix_fracdigits', 'Unit', 'physical_quantity',
           'difference_of_orders_of_magnitude',
           'Fraction', 'prime_factors',
           'ClockTime']



# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from abc import ABCMeta, abstractmethod

from mathmakerlib.calculus.number import Number
from mathmakerlib.core.drawable import Drawable


class Equilateral(Drawable, metaclass=ABCMeta):

    def __init__(self, mark_equal_sides=True, use_mark='||'):
        if mark_equal_sides:
            for s in self.sides:
                s.mark = use_mark

    @property
    @abstractmethod
    def sides(self):
        """The sides of the object."""

    @property
    def lbl_side_length(self):
        collected_values = []
        for s in self.sides:
            if isinstance(s.label_value, Number):
                collected_values.append(s.label_value)
        if not collected_values:
            raise ValueError('Found no side labeled as a Number.')
        ref = collected_values[0]
        for v in collected_values:
            if v != ref:
                raise ValueError('Found different values for the sides: {} '
                                 'and {}.'.format(repr(ref), repr(v)))
        return ref

    @property
    def default_masks(self):
        return [' ' for _ in range(len(self.sides) - 1)] + [None]



# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


class AttrList(object):

    def __init__(self, *attrlist, braces='{}'):
        if all([o is None for o in attrlist]):
            attrlist = []
        if (len(attrlist) == 1 and isinstance(attrlist[0], AttrList)):
            self._content = list(attrlist[0]._content)
            self._braces = attrlist[0]._braces
        else:
            self._content = list(attrlist)
            self._braces = braces

    def __str__(self):
        if not len(self._content):
            return ''
        built_content = []
        for o in self._content:
            if isinstance(o, dict):
                for k in o:
                    built_content.append('{}={}'.format(k, o[k]))
            else:
                built_content.append(str(o))
        return '{}{}{}'.format(self._braces[0],
                               ', '.join(built_content),
                               self._braces[1])

    def append(self, value):
        self._content.append(value)


class OptionsList(AttrList):

    def __init__(self, *options):
        super().__init__(*options, braces='[]')



# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from abc import ABCMeta, abstractmethod

from mathmakerlib.geometry.point import Point
from mathmakerlib.core.drawable import HasThickness, Colored


class Polyhedron(Colored, HasThickness, metaclass=ABCMeta):
    """Polyhedra (not drawables yet)."""

    def __init__(self, *vertices, name=None,
                 draw_vertices=False, label_vertices=True,
                 thickness='thick', color=None):
        r"""
        Initialize Polyhedron.

        :param vertices: the vertices of the Polyhedron
        :type vertices: a list of at least three Points
        :param name: the name of the Polyhedron, like ABCDEFGH.
        Can be either None (the names will be automatically created), or a
        string of the letters to use to name the vertices. Only single letters
        are supported as Points' names so far (at Polyhedron's creation).
        See issue #3.
        :type name: None or str
        :param draw_vertices: whether to actually draw, or not, the vertices
        :type draw_vertices: bool
        :param label_vertices: whether to label, or not, the vertices
        :type label_vertices: bool
        :param thickness: the thickness of the Polyhedron's edges
        :type thickness: str
        :param color: the color of the Polyhedron's edges
        :type color: str
        """
        if len(vertices) <= 3:
            raise ValueError('At least four Points are required to be able '
                             'to build a Polyhedron. Found only {} positional '
                             'arguments, though.'.format(len(vertices)))
        if any([not (isinstance(v, Point) and v.three_dimensional)
                for v in vertices]):
            for i, v in enumerate(vertices):
                if not isinstance(v, Point):
                    raise TypeError('Only Points must be provided in order to '
                                    'build a Polyhedron. Yet found {} as '
                                    'positional argument #{}.'
                                    .format(type(v), i))
                if not v.three_dimensional:
                    raise TypeError('Points used to build a Polyhedron must '
                                    'be three-dimensional Points. Found a '
                                    'two-dimensional Point instead: {}.'
                                    .format(repr(v)))
        if name is not None:
            if not isinstance(name, str):
                raise TypeError('name must be a str, found {} instead.'
                                .format(repr(name)))
            if not len(name) == len(vertices):
                raise ValueError('A polyhedron\'s name must contain as many '
                                 'letters as the polyhedron\'s number of '
                                 'vertices, yet found {} letters (name: {}) '
                                 'and {} vertices.'
                                 ''.format(len(name), repr(name),
                                           len(vertices)))
        self.thickness = thickness
        self.color = color
        self.draw_vertices = draw_vertices
        self.label_vertices = label_vertices
        self._vertices = []
        for i, v in enumerate(vertices):
            if name is None:
                vname = v.name
            else:
                vname = name[i]
            if v.name == v.label:
                lbl = 'default'
            else:
                lbl = v.label
            self._vertices.append(Point(v.x, v.y, v.z, name=vname,
                                        shape=v.shape,
                                        label=lbl, color=v.color,
                                        shape_scale=v.shape_scale))
        self._init_faces()
        self._edges = []
        for F in self.faces:
            for s in F.sides:
                if s not in self._edges:
                    self._edges.append(s)

    @property
    def vertices(self):
        return self._vertices

    @property
    def edges(self):
        return self._edges

    @property
    def faces(self):
        return self._faces

    @abstractmethod
    def _init_faces(self):
        """Each new Polyhedron must define its faces."""

    @property
    def name(self):
        return ''.join([v.name for v in self.vertices])

    @property
    def draw_vertices(self):
        return self._draw_vertices

    @draw_vertices.setter
    def draw_vertices(self, value):
        if isinstance(value, bool):
            self._draw_vertices = value
        else:
            raise TypeError('draw_vertices must be a boolean; '
                            'got {} instead.'.format(type(value)))

    @property
    def label_vertices(self):
        return self._label_vertices

    @label_vertices.setter
    def label_vertices(self, value):
        if isinstance(value, bool):
            self._label_vertices = value
        else:
            raise TypeError('label_vertices must be a boolean; '
                            'got {} instead.'.format(type(value)))



# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from . import AttrList, OptionsList


class Command(object):

    def __init__(self, name, content=None, options=None):
        self._name = str(name)
        if isinstance(content, (tuple, list)):
            self._content = AttrList(*content)
        else:
            self._content = AttrList(content)
        if isinstance(options, (tuple, list)):
            self._options = OptionsList(*options)
        else:
            self._options = OptionsList(options)

    def __str__(self):
        return r'\{name}{options}{content}'.format(name=self._name,
                                                   options=str(self._options),
                                                   content=str(self._content))

    @property
    def content(self):
        return self._content

    @property
    def options(self):
        return self._options


class DocumentClass(Command):

    def __init__(self, content=None, options=None):
        super().__init__('documentclass', content=content, options=options)


class UsePackage(Command):

    def __init__(self, content=None, options=None):
        super().__init__('usepackage', content=content, options=options)


class UseTikzLibrary(Command):

    def __init__(self, content=None, options=None):
        super().__init__('usetikzlibrary', content=content, options=options)



# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from mathmakerlib.exceptions import ZERO_OBJECTS_ERRORS, ZeroVector
from mathmakerlib.geometry.point import Point
from mathmakerlib.geometry.vector import Vector
from mathmakerlib.calculus.number import Number
from mathmakerlib.core.dimensional import Dimensional
from mathmakerlib.calculus.tools import is_number, is_integer


class Bipoint(Dimensional):
    """
    A pair of Points. Gather methods common to LineSegment, Line, Ray.

    Bipoints are quite close to, but not completely the same as, bound vectors.
    For free vectors, see Vector.

    Notice that if:
    A = Point(0, 0); B = Point(1, 0); C = Point(0, 1) and D = Point(1, 1),
    then: Bipoint(A, B) != Bipoint(C, D)
    but: Vector(A, B) == Vector(C, D)

    Also, note that contrary to LineSegments, Bipoint(A, B) != Bipoint(B, A).

    This class won't ever need to get Drawable, but can be instanciated.
    """

    def __init__(self, tail, head, allow_zero_length=True):
        """
        A Bipoint can be created from a pair of Points or a Point + a Vector.

        :param tail: the first Point of the Bipoint
        :type tail: Point
        :param head: the second Point of the Bipoint. If a Vector is provided,
        the second Point will be calculated using the first Point and this
        Vector.
        :type head: Point or Vector
        :param allow_zero_length: whether zero length Bipoints are allowed or
        not (default True).
        :type allow_zero_length: bool
        """
        if not isinstance(tail, Point):
            raise TypeError('First argument must be a Point, found {} '
                            'instead.'.format(repr(tail)))
        if not isinstance(head, (Point, Vector)):
            raise TypeError('Second argument must be a Point or a Vector, '
                            'found {} instead.'.format(repr(head)))
        self._three_dimensional = tail.three_dimensional \
            or head.three_dimensional
        if isinstance(head, Vector):
            if self._three_dimensional:
                zval = tail.z + head.z
            else:
                zval = 'undefined'
            head = Point(tail.x + head.x, tail.y + head.y, zval)
        if (not allow_zero_length
            and tail.coordinates == head.coordinates):
            msg = 'Explicitly disallowed creation of a zero-length {}.'\
                .format(type(self).__name__)
            raise ZERO_OBJECTS_ERRORS[type(self).__name__](msg)
        self._points = [tail, head]
        self._Δx = self.points[1].x - self.points[0].x
        self._Δy = self.points[1].y - self.points[0].y
        self._Δz = self.points[1].z - self.points[0].z

    def __repr__(self):
        return 'Bipoint({}, {})'.format(repr(self.tail), repr(self.head))

    def __eq__(self, other):
        if isinstance(other, Bipoint):
            return self.head == other.head and self.tail == other.tail
        else:
            return False

    def __add__(self, other):
        return self.add(other)

    def add(self, other, new_endpoint_name='automatic'):
        if not isinstance(other, Bipoint):
            raise TypeError('Can only add a Bipoint to another Bipoint. '
                            'Found {} instead.'.format(repr(other)))
        if self.three_dimensional:
            zval = self.points[1].z + other.Δz
        else:
            zval = 'undefined'
        return Bipoint(self.points[0],
                       Point(self.points[1].x + other.Δx,
                             self.points[1].y + other.Δy,
                             z=zval,
                             name=new_endpoint_name))

    @property
    def points(self):
        return self._points

    @property
    def tail(self):
        return self.points[0]

    @property
    def head(self):
        return self.points[1]

    @property
    def Δx(self):
        return self._Δx

    @property
    def Δy(self):
        return self._Δy

    @property
    def Δz(self):
        return self._Δz

    @property
    def coordinates(self):
        return (self._Δx, self._Δy, self._Δz)

    @property
    def length(self):
        """Length between the two Points."""
        return Number(self.Δx ** 2 + self.Δy ** 2 + self.Δz ** 2)\
            .sqrt()

    @property
    def slope(self):
        """Slope of the pair of Points, from -180° to 180°."""
        try:
            return Vector(self).slope
        except ZeroVector:
            msg = 'Cannot calculate the slope of a zero-length {}.'\
                .format(type(self).__name__)
            raise ZERO_OBJECTS_ERRORS[type(self).__name__](msg)

    @property
    def slope360(self):
        """Slope of the pair of Points, from 0° to 360°."""
        try:
            return Vector(self).slope360
        except ZeroVector:
            msg = 'Cannot calculate the slope of a zero-length {}.'\
                .format(type(self).__name__)
            raise ZERO_OBJECTS_ERRORS[type(self).__name__](msg)

    def midpoint(self, name='automatic'):
        """Bipoint's midpoint."""
        if self.three_dimensional:
            zval = (self.points[0].z + self.points[1].z) / 2
        else:
            zval = 'undefined'
        return Point((self.points[0].x + self.points[1].x) / 2,
                     (self.points[0].y + self.points[1].y) / 2,
                     z=zval,
                     name=name)

    def point_at(self, position, name='automatic'):
        """
        A Point aligned with the Bipoint, at provided position.

        The Bipoint's length is the length unit of position.
        Hence, position 0 matches points[0], position 1 matches points[1],
        position 0.5 matches the midpoint, position 0.75 is three quarters
        on the way from points[0] to points[1], position 2 is a Point that
        makes points[1] the middle between it and points[0], position -1 makes
        points[0] the middle between it and points[1].

        :param position: a number
        :type position: number
        :param name: the name to give to the Point
        :type name: str
        """
        if not is_number(position):
            raise TypeError('position must be a number, found {} instead.'
                            .format(type(position)))
        k = Number(position)
        if k == 0:
            return self.points[0]
        elif k == 1:
            return self.points[1]
        else:
            if self.three_dimensional:
                zval = (self.points[0].z
                        + (self.points[1].z - self.points[0].z) * k)
            else:
                zval = 'undefined'
            return Point((self.points[0].x
                          + (self.points[1].x - self.points[0].x) * k),
                         (self.points[0].y
                          + (self.points[1].y - self.points[0].y) * k),
                         z=zval,
                         name=name)

    def dividing_points(self, n=None, prefix='a'):
        """
        Create the list of Points that divide the Bipoint in n parts.

        :param n: the number of parts (so it will create n - 1 points)
        n must be greater or equal to 1
        :type n: int
        """
        if not (is_number(n) and is_integer(n)):
            raise TypeError('n must be an integer')
        if not n >= 1:
            raise ValueError('n must be greater or equal to 1')
        x0 = self.points[0].x
        x1 = self.points[1].x
        xstep = (x1 - x0) / n
        x_list = [x0 + (i + 1) * xstep for i in range(int(n - 1))]
        y0 = self.points[0].y
        y1 = self.points[1].y
        ystep = (y1 - y0) / n
        y_list = [y0 + (i + 1) * ystep for i in range(int(n - 1))]
        if self.three_dimensional:
            z0 = self.points[0].z
            z1 = self.points[1].z
            zstep = (z1 - z0) / n
            z_list = [z0 + (i + 1) * zstep for i in range(int(n - 1))]
        else:
            z_list = ['undefined' for i in range(int(n - 1))]
        return [Point(x, y, z, prefix + str(i + 1))
                for i, (x, y, z) in enumerate(zip(x_list, y_list, z_list))]



# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from mathmakerlib import config
from mathmakerlib.calculus.number import Number
from mathmakerlib.geometry.point import Point
from . import Triangle


class IsoscelesTriangle(Triangle):
    """Isosceles Triangles."""

    def __init__(self, start_vertex=None, name=None,
                 base_length=Number('1.5'),
                 equal_legs_length=Number('1'),
                 mark_equal_sides=True, use_mark='||',
                 draw_vertices=False, label_vertices=True,
                 thickness='thick', color=None, rotation_angle=0,
                 winding=None, sloped_sides_labels=True):
        r"""
        Initialize Isosceles Triangle

        :param start_vertex: the vertex to start to draw the Right Triangle
        (default (0; 0))
        :type start_vertex: Point
        :param name: the name of the Triangle, like ABC.
        Can be either None (the names will be automatically created), or a
        string of the letters to use to name the vertices. Only single letters
        are supported as Points' names so far (at Polygon's creation).
        See issue #3.
        :type name: None or str
        :param base_length: the length of the base of the IsoscelesTriangle,
        that will be used to calculate the coordinates of its vertices
        :type base_length: a number
        :param equal_legs_length: the length of the equal legs of the
        IsoscelesTriangle, that will be used to calculate the coordinates of
        its vertices
        :type equal_legs_length: a number
        :param mark_equal_sides: if True (default), all three sides will be
        automatically marked with the same symbol.
        :type mark_equal_sides: bool
        :param draw_vertices: whether to actually draw, or not, the vertices
        :type draw_vertices: bool
        :param label_vertices: whether to label, or not, the vertices
        :type label_vertices: bool
        :param thickness: the thickness of the Triangle's sides
        :type thickness: str
        :param color: the color of the Triangle's sides
        :type color: str
        :param rotate: the angle of rotation around isobarycenter
        :type rotate: int
        """
        if start_vertex is None:
            start_vertex = Point(0, 0)
        self._base_length = Number(base_length)
        self._equal_legs_length = Number(equal_legs_length)
        v1 = Point(base_length + start_vertex.x, start_vertex.y)
        v2 = Point(base_length / 2,
                   (equal_legs_length ** 2 - Number('0.25') * base_length ** 2)
                   .sqrt().rounded(Number('0.001')) + start_vertex.y)
        if (winding == 'clockwise'
            or (winding is None
                and config.polygons.DEFAULT_WINDING == 'clockwise')):
            start_vertex, v1 = v1, start_vertex
        Triangle.__init__(self, start_vertex, v1, v2, name=name,
                          draw_vertices=draw_vertices,
                          label_vertices=label_vertices,
                          thickness=thickness, color=color,
                          rotation_angle=rotation_angle,
                          winding=winding,
                          sloped_sides_labels=sloped_sides_labels)
        self._type = 'IsoscelesTriangle'
        if mark_equal_sides:
            self.sides[1].mark = self.sides[2].mark = use_mark

    @property
    def base_length(self):
        return self._base_length

    @property
    def equal_legs_length(self):
        return self._equal_legs_length

    def setup_labels(self, labels=None, linesegments=None, masks=None):
        """
        Convenience to easily setup IsoscelesTriangle's sides length labels.

        If masks is None, then by default, only base's label and sides[2].label
        will be shown. The other will be masked.

        If labels has only two elements, then the second one will be considered
        as the length of equal length sides and will be duplicated.

        :param labels: None or the list of the labels
        :type labels: None or list of 2 or 3 elements
        :param linesegments: the list of the LineSegments to label
        (defaults to Polygon's sides)
        :type linesegments: list (of LineSegments)
        :param masks: the list of masks to setup.
        :type masks: None or list of 3 elements
        """
        if masks is None:
            if self.winding == 'anticlockwise':
                masks = [None, ' ', None]
            else:
                masks = [None, None, ' ']
        if len(labels) == 2:
            labels = [labels[0], labels[1], labels[1]]
        if linesegments is None:
            linesegments = self.sides
        super().setup_labels(labels=labels, masks=masks,
                             linesegments=linesegments)



# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import sys

LOCALE_US = 'en' if sys.platform.startswith('win') else 'en_US.UTF-8'
LOCALE_FR = 'fr' if sys.platform.startswith('win') else 'fr_FR.UTF-8'



# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from mathmakerlib.LaTeX import DASHPATTERN_VALUES
from mathmakerlib.exceptions import ZeroLengthLineSegment
from mathmakerlib.core.drawable import check_scale, Drawable, HasThickness
from mathmakerlib.core.drawable import tikz_approx_position, tikz_options_list
from mathmakerlib.geometry.bipoint import Bipoint
from mathmakerlib.calculus.number import Number
from mathmakerlib.geometry.point import OPPOSITE_LABEL_POSITIONS

LABEL_MASK_VALUES = [None, ' ', '?']


class LineSegment(Drawable, HasThickness, Bipoint):

    def __init__(self, *points, thickness='thick', dashpattern='solid',
                 label=None, label_mask=None, label_winding='anticlockwise',
                 label_position=None, label_scale=None, mark=None,
                 mark_scale=Number('0.5'), color=None,
                 draw_endpoints=True, label_endpoints=True,
                 locked_label=False, allow_zero_length=True,
                 sloped_label=True):
        """
        Initialize LineSegment

        :param points: either one LineSegment to copy, or two Points
        :type points: another LineSegment or a list or tuple of two Points
        :param thickness: the LineSegment's thickness. Available values are
        TikZ's ones.
        :type thickness: str
        :param label: what will be written along the LineSegment, about its
        middle. A None value will disable the LineSegment's labeling.
        :type label: None or str
        :param label_mask: if not None (default), hide the label with ' '
        or '?'
        :type label_mask: None or str (' ' or '?')
        :param label_position: tells where to put the LineSegment's label.
        Can be a value used by TikZ or 'clockwise' or 'anticlockwise'.
        'anticlockwise' (default) will automatically set the label_position to
        'above' if Δx < 0 and to 'below' if Δx > 0. This is useful to
        put all LineSegments' labels outside a Polygon that is drawn in an
        anticlockwise manner. Same for 'clockwise', in the reversed direction.
        :type label_position: str
        :param label_scale: the label's scale
        :type label_scale: None or anything that is accepted to create a Number
        :param draw_endpoints: whether or not actually draw the endpoints.
        Defaults to True.
        :type draw_endpoints: bool
        :param label_endpoints: whether or not label the endpoints.
        Defaults to True.
        :type label_endpoints: bool
        :param mark: the mark to print on the line segment
        :type mark: str
        :param mark_scale: the scale (size) of the mark. Defaults to 0.5
        :type mark_scale: any number
        :param locked_label: to allow or prevent, by default, modifications of
        the LineSegment's label.
        :type locked_label: bool
        """
        if len(points) != 2:
            raise TypeError('Two Points are required to create a '
                            'LineSegment. Got {} object(s) instead.'
                            .format(len(points)))
        Bipoint.__init__(self, *points, allow_zero_length=allow_zero_length)
        self._label = None
        self._thickness = None
        self._label_mask = None
        self._label_scale = None
        self._draw_endpoints = None
        self._label_endpoints = None
        self._locked_label = False  # Only temporary, in order to be able to
        self.label = label          # use the label setter.
        self.label_mask = label_mask
        self.label_scale = label_scale
        self.thickness = thickness
        self.dashpattern = dashpattern
        self.draw_endpoints = draw_endpoints
        self.sloped_label = sloped_label
        self.label_winding = label_winding
        self.label_endpoints = label_endpoints
        self.mark = mark
        self.mark_scale = mark_scale
        try:
            self.endpoints[1].label_position = \
                tikz_approx_position(self.slope360)
        except ZeroLengthLineSegment:
            self.endpoints[1].label_position = 'below left'
        self.endpoints[0].label_position = \
            OPPOSITE_LABEL_POSITIONS[self.endpoints[1].label_position]
        if label_position is None:
            label_position = 'automatic'
        self.label_position = label_position
        if color is not None:
            self.color = color
        self._comment_designation = 'Line Segment'
        if not isinstance(locked_label, bool):
            raise TypeError('Expected bool type for \'locked_label\' keyword '
                            'argument. Found {}.'.format(type(locked_label)))
        self._locked_label = locked_label

    def __repr__(self):
        return 'LineSegment({}, {})'.format(repr(self.endpoints[0]),
                                            repr(self.endpoints[1]))

    def __eq__(self, other):
        if isinstance(other, LineSegment):
            return ((self.tail == other.tail and self.head == other.head)
                    or (self.tail == other.head and self.head == other.tail))
        else:
            return False

    def __hash__(self):
        points = sorted(set(self.endpoints),
                        key=lambda point: point.coordinates)
        if self.three_dimensional:
            s = 'Point ({}, {}, {}) - Point ({}, {}, {})'\
                .format(*points[0].coordinates, *points[1].coordinates)
        else:
            s = 'Point ({}, {}) - Point ({}, {})'\
                .format(points[0].x, points[0].y, points[1].x, points[1].y)
        return hash(s)

    @property
    def dashpattern(self):
        return self._dashpattern

    @dashpattern.setter
    def dashpattern(self, value):
        if value in DASHPATTERN_VALUES:
            self._dashpattern = value
        else:
            raise ValueError('Incorrect dashpattern value: \'{}\'. '
                             'Available values belong to: {}.'
                             .format(str(value), str(DASHPATTERN_VALUES)))

    @property
    def label_endpoints(self):
        return self._label_endpoints

    @label_endpoints.setter
    def label_endpoints(self, value):
        if value in [True, False]:
            self._label_endpoints = value
        else:
            raise ValueError('label_endpoints must be True or False; '
                             'got \'{}\' instead.'.format(value))

    @property
    def draw_endpoints(self):
        return self._draw_endpoints

    @draw_endpoints.setter
    def draw_endpoints(self, value):
        if value in [True, False]:
            self._draw_endpoints = value
        else:
            raise ValueError('draw_endpoints must be True or False; '
                             'got \'{}\' instead.'.format(value))

    @property
    def endpoints(self):
        return self._points

    # @property
    # def name(self):
    #     return '[' + self.endpoints[0].name + self.endpoints[1].name + ']'

    @property
    def locked_label(self):
        """Tell if the LineSegment's label can be changed."""
        return self._locked_label

    def lock_label(self):
        """Forbid modifications of LineSegment's label."""
        self._locked_label = True

    def unlock_label(self):
        """Allow modifications of LineSegment's label."""
        self._locked_label = False

    @property
    def label_winding(self):
        return self._label_winding

    @label_winding.setter
    def label_winding(self, value):
        if value in ['clockwise', 'anticlockwise']:
            self._label_winding = value
        else:
            raise ValueError("label_winding must be 'clockwise' or "
                             "'anticlockwise'; found {} instead."
                             .format(repr(value)))

    @property
    def sloped_label(self):
        return self._sloped_label

    @sloped_label.setter
    def sloped_label(self, value):
        if value in [True, False]:
            self._sloped_label = value
        else:
            raise ValueError('sloped_label must be True or False; '
                             'got \'{}\' instead.'.format(value))

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, value):
        if self.locked_label:
            raise TypeError('This LineSegments\' label is locked. '
                            'If you\'re using this LineSegment embedded in '
                            'another object (e.g. a Polygon), please use the '
                            'setup_labels() method of this object. Otherwise, '
                            'first explicitely unlock the LineSegment.')
        else:
            super(LineSegment, self.__class__).label.fset(self, value)

    @property
    def label_position(self):
        return self._label_position

    @label_position.setter
    def label_position(self, value):
        if self.sloped_label:
            if value == 'automatic':
                if self.label_winding == 'anticlockwise':
                    if self.Δx >= 0:
                        self._label_position = 'below'
                    else:
                        self._label_position = 'above'
                elif self.label_winding == 'clockwise':
                    if self.Δx >= 0:
                        self._label_position = 'above'
                    else:
                        self._label_position = 'below'
            else:
                self._label_position = str(value)
        else:
            if value == 'automatic':
                ω = {'anticlockwise': -90, 'clockwise': 90}[self.label_winding]
                self._label_position = tikz_approx_position(self.slope360 + ω)
            else:
                self._label_position = str(value)

    @property
    def label_mask(self):
        return self._label_mask

    @label_mask.setter
    def label_mask(self, value):
        if value in LABEL_MASK_VALUES:
            self._label_mask = value
        else:
            raise ValueError('label_mask must be in {}; got \'{}\' instead.'
                             .format(LABEL_MASK_VALUES, value))

    @property
    def mark(self):
        return self._mark

    @mark.setter
    def mark(self, value):
        if value is not None:
            self._mark = str(value)
        else:
            self._mark = None

    @property
    def mark_scale(self):
        return self._mark_scale

    @mark_scale.setter
    def mark_scale(self, value):
        check_scale(value, 'LineSegment\'s mark')
        self._mark_scale = Number(value)

    @property
    def label_scale(self):
        return self._label_scale

    @label_scale.setter
    def label_scale(self, value):
        if value is None:
            self._label_scale = None
        else:
            self._label_scale = Number(value)

    def tikz_declarations(self):
        """Return the LineSegment's Points' declarations."""
        return '{}\n{}'\
            .format(self.endpoints[0].tikz_declarations(),
                    self.endpoints[1].tikz_declarations())

    def tikz_drawing_comment(self):
        """Return the comments preceding the LineSegment's drawings."""
        ls_draw_comment = '% Draw {}'.format(self._comment_designation)
        if self.draw_endpoints:
            return ['% Draw Points', ls_draw_comment]
        else:
            return [ls_draw_comment]

    def _tikz_draw_options(self):
        output = [self.thickness, self.color]
        if self.dashpattern != 'solid':
            output.append(self.dashpattern)
        return output

    def _tikz_draw_endpoints(self):
        if self.draw_endpoints:
            return ['{}\n{}\n'.format(self.endpoints[0].tikz_draw()[0],
                                      self.endpoints[1].tikz_draw()[0])]
        else:
            return []

    def _tikz_label_options(self):
        options = ['midway', self.label_position]
        if self.sloped_label:
            options.append('sloped')
        if self.label_scale is not None:
            options.append('scale={}'.format(self.label_scale))
        return options

    def tikz_label(self):
        lbl = ''
        if self.label_mask is None and self.label is not None:
            lbl = self.label
        elif self.label_mask is not None and self.label_mask != ' ':
            lbl = self.label_mask
        if lbl != '':
            return ' node{} {}'\
                .format(tikz_options_list(self._tikz_label_options()),
                        '{' + lbl + '}')
        else:
            return lbl

    def _tikz_ls_mark(self):
        if self.mark is not None:
            return ' node[midway, sloped, scale={}] {}'\
                .format(self.mark_scale, '{' + self.mark + '}')
        else:
            return ''

    def tikz_draw_section(self):
        return '-- ({}){}{}'.format(self.endpoints[1].name,
                                    self.tikz_label(),
                                    self._tikz_ls_mark())

    def tikz_draw(self):
        """Return the command to actually draw the LineSegment."""
        output = self._tikz_draw_endpoints()
        output.append(r'\draw{} ({}) {};'
                      .format(tikz_options_list('draw', self),
                              self.endpoints[0].name,
                              self.tikz_draw_section()))
        return output

    def tikz_points_labels(self):
        """Return the command to write the labels."""
        output = ''
        if self.label_endpoints:
            output = '{}\n{}'.format(self.endpoints[0].tikz_label(),
                                     self.endpoints[1].tikz_label())
        return output



# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

"""
Exponented objects are any calculable objects that have an exponent, like in
7³ or x².
"""


import copy

from mathmakerlib.core.printable import Printable
from mathmakerlib.core.signed import Signed


class Exponented(Printable):
    """
    Exponented objects are any calculable objects that have an exponent, like
    in 7³ or x². The lower part is called content, the higher part is the
    exponent.
    """

    def __init__(self, content, exponent=None):
        self._content = None
        self.content = content
        self._exponent = None
        self.exponent = exponent

    def __repr__(self):
        if self.exponent is None:
            return 'Exponented({})'.format(repr(self.content))
        else:
            return 'Exponented({}^{})'.format(repr(self.content),
                                              repr(self.exponent))

    def __str__(self):
        if self.exponent is not None:
            return str(self.content) + '^' + str(self.exponent)
        else:
            return str(self.content)

    @property
    def content(self):
        """The lower part of the Exponented object."""
        return self._content

    @content.setter
    def content(self, content):
        if isinstance(content, int):
            from mathmakerlib.calculus import Number
            content = Number(content)
        if not isinstance(content, Printable):
            raise TypeError('The content of an Exponented must be a Printable '
                            'object. An int would be '
                            'accepted (turned into Number). Got {} instead.'
                            .format(str(type(content))))
        self._content = copy.deepcopy(content)

    @property
    def exponent(self):
        """The exponent part of the Exponented object."""
        return self._exponent

    @exponent.setter
    def exponent(self, exponent):
        if isinstance(exponent, int):
            from mathmakerlib.calculus import Number
            exponent = Number(exponent)
        if not (exponent is None or isinstance(exponent, Printable)):
            raise TypeError('The exponent of an Exponented must be either '
                            'None or a Printable object. An int would be '
                            'accepted (turned into Number). Got {} instead.'
                            .format(str(type(exponent))))
        self._exponent = copy.deepcopy(exponent)

    def imprint(self, start_expr=True, variant='latex'):
        if self.exponent is None:
            return self.content.imprint(start_expr=start_expr, variant=variant)
        else:
            open_brace = close_brace = ''
            if isinstance(self.content, Signed) and self.content.sign == '-':
                open_brace = '('
                close_brace = ')'
            open_curly = '{'
            close_curly = '}'
            if variant == 'user_input':
                open_curly = close_curly = ''
            return open_brace \
                + self.content.imprint(start_expr=start_expr, variant=variant)\
                + close_brace + '^' + open_curly \
                + self.exponent.imprint(start_expr=True, variant=variant) \
                + close_curly



# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from decimal import Decimal


def is_number(n):
    """Check if n is a number."""
    return isinstance(n, (float, int, Decimal))


def is_integer(n):
    """Check if number n is an integer."""
    if isinstance(n, int):
        return True
    elif isinstance(n, float):
        return n.is_integer()
    elif isinstance(n, Decimal):
        return n % 1 == 0
    else:
        raise TypeError('Expected a number, either float, int or Decimal,'
                        'got {} instead.'.format(str(type(n))))


def is_natural(n):
    """Check if number n is a natural number."""
    return is_integer(n) and n >= 0


def prime_factors(n):
    """
    Return all the prime factors of a positive integer

    Taken from https://stackoverflow.com/a/412942/3926735.
    """
    try:
        n = int(n)
    except ValueError:
        raise TypeError('The argument must be an integer.')
    factors = []
    d = 2
    while n > 1:
        while n % d == 0:
            factors.append(d)
            n //= d
        d = d + 1
        if d * d > n:
            if n > 1:
                factors.append(n)
            break
    return factors



# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# To change default mathmakerlib's settings:

# import mathmakerlib.config
# mathmakerlib.config.polygons.DEFAULT_WINDING = 'clockwise'

from decimal import Decimal

from mathmakerlib.LaTeX import DASHPATTERN_VALUES
from mathmakerlib.core.oriented import check_winding
from mathmakerlib.calculus.number import Number, is_number
from mathmakerlib.geometry.projections.oblique_projection \
    import DIRECTION_VALUES
from mathmakerlib.calculus.clocktime import check_clocktime_context
from mathmakerlib.calculus.clocktime import DEFAULT_CLOCKTIME_CONTEXT


class PolygonsSetup(object):

    def __init__(self):
        # default_winding can be set to either 'clockwise', 'anticlockwise' or
        # None. If it is left to None, the polygons' winding won't be forced
        # to a default value (if not overriden by the user at Polygon's
        # initialization), but deduced from the given vertices' order.
        # Use with caution: when you force the winding to be 'clockwise' by
        # default and give anticlockwise oriented vertices to build a Polygon,
        # the vertices' names' order in the final created Polygon WILL NOT be
        # the same as the given one (it cannot be the same). Yet setup_labels()
        # and setup_marks() WILL remember the order you gave the vertices to
        # create the Polygon (this places the marks and labels at expected
        # places...). By default, a warning will raise if the winding is forced
        # to the opposite warning of given Points when building a Polygon.
        self.DEFAULT_WINDING = None
        self.ENABLE_MISMATCH_WINDING_WARNING = True

    @property
    def DEFAULT_WINDING(self):
        return self._DEFAULT_WINDING

    @DEFAULT_WINDING.setter
    def DEFAULT_WINDING(self, value):
        if value is not None:
            check_winding(value)
        self._DEFAULT_WINDING = value

    @property
    def ENABLE_MISMATCH_WINDING_WARNING(self):
        return self._ENABLE_MISMATCH_WINDING_WARNING

    @ENABLE_MISMATCH_WINDING_WARNING.setter
    def ENABLE_MISMATCH_WINDING_WARNING(self, value):
        if value:
            self._ENABLE_MISMATCH_WINDING_WARNING = True
        else:
            self._ENABLE_MISMATCH_WINDING_WARNING = False


class PointsSetup(object):

    def __init__(self):
        self.DEFAULT_POSITION_PRECISION = Decimal('1.000')

    @property
    def DEFAULT_POSITION_PRECISION(self):
        return self._DEFAULT_POSITION_PRECISION

    @DEFAULT_POSITION_PRECISION.setter
    def DEFAULT_POSITION_PRECISION(self, value):
        if not is_number(value):
            raise TypeError('DEFAULT_POSITION_PRECISION must be a number, '
                            'found {} instead.'.format(type(value)))
        self._DEFAULT_POSITION_PRECISION = value


class AnglesSetup(object):

    def __init__(self):
        self.DEFAULT_ARMSPOINTS_POSITION = Number('0.8')

    @property
    def DEFAULT_ARMSPOINTS_POSITION(self):
        return self._DEFAULT_ARMSPOINTS_POSITION

    @DEFAULT_ARMSPOINTS_POSITION.setter
    def DEFAULT_ARMSPOINTS_POSITION(self, value):
        if not is_number(value):
            raise TypeError('DEFAULT_ARMSPOINTS_POSITION must be a number, '
                            'found {} instead.'.format(type(value)))
        self._DEFAULT_ARMSPOINTS_POSITION = value


class ClockTimeSetup(object):

    def __init__(self):
        self._CONTEXT = DEFAULT_CLOCKTIME_CONTEXT

    @property
    def CONTEXT(self):
        return self._CONTEXT

    @CONTEXT.setter
    def CONTEXT(self, context):
        check_clocktime_context(context)
        self._CONTEXT.update(context)


class ObliqueProjectionSetup(object):

    def __init__(self):
        self.RECEDING_AXIS_ANGLE = Number(45)
        self.RATIO = Number(0.67)
        self.DASHPATTERN = 'dashed'
        self.DIRECTION = 'top-right'

    @property
    def RECEDING_AXIS_ANGLE(self):
        return self._RECEDING_AXIS_ANGLE

    @RECEDING_AXIS_ANGLE.setter
    def RECEDING_AXIS_ANGLE(self, value):
        if not is_number(value):
            raise TypeError('RECEDING_AXIS_ANGLE must be a number, '
                            'found {} instead.'.format(repr(value)))
        self._RECEDING_AXIS_ANGLE = value

    @property
    def RATIO(self):
        return self._RATIO

    @RATIO.setter
    def RATIO(self, value):
        if not is_number(value):
            raise TypeError('RATIO must be a number, found {} instead.'
                            .format(repr(value)))
        self._RATIO = value

    @property
    def DASHPATTERN(self):
        return self._DASHPATTERN

    @DASHPATTERN.setter
    def DASHPATTERN(self, value):
        if value not in DASHPATTERN_VALUES:
            raise ValueError('Incorrect dashpattern value: \'{}\'. '
                             'Available values belong to: {}.'
                             .format(str(value), str(DASHPATTERN_VALUES)))
        self._DASHPATTERN = value

    @property
    def DIRECTION(self):
        return self._DIRECTION

    @DIRECTION.setter
    def DIRECTION(self, value):
        if value in DIRECTION_VALUES:
            self._direction = value
        else:
            raise ValueError('Incorrect direction value: \'{}\'. '
                             'Available values belong to: {}.'
                             .format(str(value), str(DIRECTION_VALUES)))
        self._DIRECTION = value


SUPPORTED_LANGUAGES = ['en', 'en_US', 'en_GB', 'fr', 'fr_FR']


def init():
    global polygons, angles, oblique_projection, language, points, clocktime
    global initialized

    try:
        initialized
    except NameError:
        initialized = False

    if not initialized:
        initialized = True
        polygons = PolygonsSetup()
        points = PointsSetup()
        angles = AnglesSetup()
        oblique_projection = ObliqueProjectionSetup()
        clocktime = ClockTimeSetup()
        language = 'en'



# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from mathmakerlib.geometry.point import Point
from mathmakerlib.geometry.polygons.rectangle import Rectangle
from mathmakerlib.geometry.polyhedra.polyhedron import Polyhedron


class RightCuboid(Polyhedron):
    """Right Cuboids."""

    def __init__(self, start_vertex=None, dimensions=None, name=None,
                 draw_vertices=False, label_vertices=True,
                 thickness='thick', color=None):
        r"""
        Initialize Right Cuboid

        :param start_vertex: the vertex to start to draw the Right Cuboid
        (default (0; 0))
        :type start_vertex: Point (must be a 3D Point)
        :param dimensions: (width, depth, height)
        :type dimensions: tuple (of numbers)
        :param name: the name of the RightCuboid, like ABCDEFGH.
        Can be either None (the names will be automatically created), or a
        string of the letters to use to name the vertices. Only single letters
        are supported as Points' names so far (at Polyhedra's creation).
        See issue #3.
        :type name: None or str
        :param draw_vertices: whether to actually draw, or not, the vertices
        :type draw_vertices: bool
        :param label_vertices: whether to label, or not, the vertices
        :type label_vertices: bool
        :param thickness: the thickness of the RightCuboid's edges
        :type thickness: str
        :param color: the color of the RightCuboid's edges
        :type color: str
        """
        if start_vertex is None:
            start_vertex = Point(0, 0, 0)
        if not isinstance(start_vertex, Point):
            raise TypeError('start_vertex must be a Point; found {} instead.'
                            .format(repr(start_vertex)))
        if not start_vertex.three_dimensional:
            raise TypeError('start_vertex must be a three-dimensional Point. '
                            'Found this two-dimensional Point instead: {}.'
                            .format(repr(start_vertex)))
        if not isinstance(dimensions, (tuple, list)):
            raise TypeError('dimensions must be a tuple or a list. Found {} '
                            'instead.'.format(dimensions))
        if not len(dimensions) == 3:
            raise TypeError('dimensions must have a length of 3. Found {} '
                            'instead.'.format(dimensions))
        x, y, z = start_vertex.coordinates
        width, depth, height = dimensions
        vertices = [start_vertex,
                    Point(x + width, y, z),
                    Point(x + width, y + depth, z),
                    Point(x, y + depth, z),
                    Point(x, y, z + height),
                    Point(x + width, y, z + height),
                    Point(x + width, y + depth, z + height),
                    Point(x, y + depth, z + height)]
        Polyhedron.__init__(self, *vertices, name=name,
                            draw_vertices=draw_vertices,
                            label_vertices=label_vertices,
                            thickness=thickness, color=color)
        self._width, self._depth, self._height = width, depth, height
        self._labels = None
        self._edges_to_label = {}

    def _init_faces(self):
        """Faces of the RightCuboid."""
        self._faces = [Rectangle(self.vertices[0], self.vertices[1],
                                 self.vertices[2], self.vertices[3]),
                       Rectangle(self.vertices[0], self.vertices[1],
                                 self.vertices[5], self.vertices[4]),
                       Rectangle(self.vertices[1], self.vertices[2],
                                 self.vertices[6], self.vertices[5]),
                       Rectangle(self.vertices[2], self.vertices[3],
                                 self.vertices[7], self.vertices[6]),
                       Rectangle(self.vertices[3], self.vertices[0],
                                 self.vertices[4], self.vertices[7]),
                       Rectangle(self.vertices[4], self.vertices[5],
                                 self.vertices[6], self.vertices[7]),
                       ]

    def setup_labels(self, labels=None):
        """
        Easily setup RightCuboid's width, depth and height labels.

        :param labels: the labels to store: (width, depth, height)
        :type labels: tuple or list
        """
        if not (isinstance(labels, (tuple, list)) and len(labels) == 3):
            raise TypeError('labels argument must be a list or tuple, '
                            'of length 3. Found {} instead.'
                            .format(repr(labels)))
        self._labels = labels
        # These coordinates represent the face number, then the edge number.
        # Depends on _init_faces()
        self._edges_to_label = {'oblique_projection:top-right':
                                [(0, 0, 'anticlockwise'),
                                 (1, 1, 'anticlockwise'),
                                 (2, 2, 'clockwise')],
                                'oblique_projection:top-left':
                                [(0, 0, 'anticlockwise'),
                                 (1, 3, 'anticlockwise'),
                                 (4, 2, 'clockwise')],
                                'oblique_projection:bottom-left':
                                [(0, 2, 'anticlockwise'),
                                 (3, 1, 'anticlockwise'),
                                 (4, 2, 'clockwise')],
                                'oblique_projection:bottom-right':
                                [(0, 2, 'anticlockwise'),
                                 (2, 1, 'clockwise'),
                                 (2, 2, 'clockwise')],
                                }

    @property
    def width(self):
        return self._width

    @property
    def depth(self):
        return self._depth

    @property
    def height(self):
        return self._height

    @property
    def labels(self):
        return self._labels

    @property
    def lbl_width(self):
        if self.labels is None:
            raise AttributeError('Labels must be set before trying to get '
                                 'them.')
        return self.labels[0]

    @property
    def lbl_depth(self):
        if self.labels is None:
            raise AttributeError('Labels must be set before trying to get '
                                 'them.')
        return self.labels[1]

    @property
    def lbl_height(self):
        if self.labels is None:
            raise AttributeError('Labels must be set before trying to get '
                                 'them.')
        return self.labels[2]

    @property
    def edges_to_label(self):
        return self._edges_to_label

    @property
    def lbl_volume(self):
        if self.labels is None:
            raise AttributeError('Labels must be set before trying to get '
                                 'them.')
        return self.lbl_width * self.lbl_depth * self.lbl_height



# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from .drawable import Drawable
from .oriented import Oriented
from .evaluable import Evaluable
from .dimensional import Dimensional
from .printable import Printable
from .signed import Signed
from .word import Word

__all__ = ['Drawable', 'Oriented', 'Evaluable', 'Dimensional', 'Printable',
           'Signed', 'Word']



# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from math import cos, sin, radians

from mathmakerlib.calculus.tools import is_number
from mathmakerlib.calculus.number import Number
from mathmakerlib.geometry.point import Point
from . import Quadrilateral, Equilateral


class Rhombus(Quadrilateral, Equilateral):
    """Rhombi."""

    def __init__(self, start_vertex=None, name=None,
                 side_length=Number('1'),
                 build_angle=60,
                 mark_equal_sides=True, use_mark='||',
                 draw_vertices=False, label_vertices=True,
                 thickness='thick', color=None, rotation_angle=0,
                 winding=None, sloped_sides_labels=True):
        r"""
        Initialize Rhombus

        :param start_vertex: the vertex to start to draw the Rhombus
        (default (0; 0))
        :type start_vertex: Point
        :param name: the name of the Rhombus, like ABCD.
        Can be either None (the names will be automatically created), or a
        string of the letters to use to name the vertices. Only single letters
        are supported as Points' names so far (at Polygon's creation).
        See issue #3.
        :type name: None or str
        :param side_length: the length that will be used to calculate the
        coordinates of the vertices used to build the Rhombus
        :type side_length: a number
        :param build_angle: one of the interior angles of the Rhombus.
        :type build_angle: any number
        :param mark_equal_sides: if True (default), all four sides will be
        automatically marked with the same symbol.
        :type mark_equal_sides: bool
        :param draw_vertices: whether to actually draw, or not, the vertices
        :type draw_vertices: bool
        :param label_vertices: whether to label, or not, the vertices
        :type label_vertices: bool
        :param thickness: the thickness of the Quadrilateral's sides
        :type thickness: str
        :param color: the color of the Quadrilateral's sides
        :type color: str
        :param rotate: the angle of rotation around isobarycenter
        :type rotate: int
        """
        if start_vertex is None:
            start_vertex = Point(0, 0)
        self._side_length = Number(side_length)
        if is_number(build_angle):
            self._build_angle = build_angle
        else:
            raise TypeError('Expected an integer as build_angle, found {}.'
                            .format(type(build_angle)))
        x = (side_length * Number(str(cos(radians(build_angle / 2)))))\
            .rounded(Number('0.001'))
        y = (side_length * Number(str(sin(radians(build_angle / 2)))))\
            .rounded(Number('0.001'))
        v1 = Point(x + start_vertex.x, -y + start_vertex.y)
        v2 = Point(2 * x + start_vertex.x, start_vertex.y)
        v3 = Point(x + start_vertex.x, y + start_vertex.y)
        Quadrilateral.__init__(self, start_vertex, v1, v2, v3, name=name,
                               draw_vertices=draw_vertices,
                               label_vertices=label_vertices,
                               thickness=thickness, color=color,
                               rotation_angle=rotation_angle,
                               winding=winding,
                               sloped_sides_labels=sloped_sides_labels)
        self._type = 'Rhombus'
        Equilateral.__init__(self, mark_equal_sides=mark_equal_sides,
                             use_mark=use_mark)

    @property
    def side_length(self):
        return self._side_length

    def setup_labels(self, labels=None, linesegments=None, masks=None):
        """
        Convenience method to easily setup Rhombus' side length label.

        If masks is None, then by default, only sides[3].label will be shown.
        The three others will be masked.

        If labels has only one element, then it is duplicated four times.

        :param labels: None or the list of the labels
        :type labels: None or list of 1 or 4 elements
        :param linesegments: the list of the LineSegments to label
        (defaults to Polygon's sides)
        :type linesegments: list (of LineSegments)
        :param masks: the list of masks to setup.
        :type masks: None or list of 4 elements
        """
        if masks is None:
            masks = self.default_masks
        if linesegments is None:
            linesegments = self.sides
        if len(labels) == 1:
            labels *= 4
        super().setup_labels(labels=labels, linesegments=linesegments,
                             masks=masks)



# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

"""Constants and LaTeX-related stuff."""


from .attr_list import AttrList, OptionsList
from .commands import Command, DocumentClass, UsePackage, UseTikzLibrary

DEFAULT_FONT_SIZES = [r'\tiny', r'\scriptsize', r'\footnotesize', r'\small',
                      r'\normalsize', r'\large', r'\Large', r'\LARGE',
                      r'\huge', r'\Huge']

DEFAULT_COLOR_NAMES = ['white', 'black', 'red', 'green', 'blue', 'cyan',
                       'magenta', 'yellow']

# Colors' names from xcolor package
XCOLOR_BASE = ['black', 'blue', 'brown', 'cyan', 'darkgray', 'gray', 'green',
               'lightgray', 'lime', 'magenta', 'olive', 'orange', 'pink',
               'purple', 'red', 'teal', 'violet', 'white', 'yellow']
XCOLOR_DVIPSNAMES = ['Apricot', 'Aquamarine', 'Bittersweet', 'Black', 'Blue',
                     'BlueGreen', 'BlueViolet', 'BrickRed', 'Brown',
                     'BurntOrange', 'CadetBlue', 'CarnationPink', 'Cerulean',
                     'CornflowerBlue', 'Cyan', 'Dandelion', 'DarkOrchid',
                     'Emerald', 'ForestGreen', 'Fuchsia', 'Goldenrod', 'Gray',
                     'Green', 'GreenYellow', 'JungleGreen', 'Lavender',
                     'LimeGreen', 'Magenta', 'Mahogany', 'Maroon', 'Melon',
                     'MidnightBlue', 'Mulberry', 'NavyBlue', 'OliveGreen',
                     'Orange', 'OrangeRed', 'Orchid', 'Peach', 'Periwinkle',
                     'PineGreen', 'Plum', 'ProcessBlue', 'Purple', 'RawSienna',
                     'Red', 'RedOrange', 'RedViolet', 'Rhodamine', 'RoyalBlue',
                     'RoyalPurple', 'RubineRed', 'Salmon', 'SeaGreen', 'Sepia',
                     'SkyBlue', 'SpringGreen', 'Tan', 'TealBlue', 'Thistle',
                     'Turquoise', 'Violet', 'VioletRed', 'White',
                     'WildStrawberry', 'Yellow', 'YellowGreen', 'YellowOrange']

TIKZSET = {'singledash_hatchmark': r"""
\tikzset{singledash/.style={decoration={ markings, mark= at position 0.5
with { \draw (0pt,-2.5pt) -- (0pt,2.5pt);
} }, pic actions/.append code=\tikzset{postaction=decorate}}}""",
           'doubledash_hatchmark': r"""
\tikzset{doubledash/.style={decoration={ markings, mark= at position 0.5
with { \draw (-1pt,-2.5pt) -- (-1pt,2.5pt);
       \draw (1pt,-2.5pt) -- (1pt,2.5pt);
} }, pic actions/.append code=\tikzset{postaction=decorate}}}""",
           'tripledash_hatchmark': r"""
\tikzset{tripledash/.style={decoration={ markings, mark= at position 0.5
with { \draw (-2pt,-2.5pt) -- (-2pt,2.5pt);
       \draw (0pt,-2.5pt) -- (0pt,2.5pt);
       \draw (2pt,-2.5pt) -- (2pt,2.5pt);
} }, pic actions/.append code=\tikzset{postaction=decorate}}}"""
           }

MATHEMATICAL_NOTATIONS = \
    {'en': {'angle_name': r'\angle {content}'},
     'fr': {'angle_name': r'\stackon[-5pt]{{{content}}}{{\vstretch{{1.5}}'
                          r'{{\hstretch{{1.6}}{{\widehat{{\phantom{{\;\;\;\;'
                          r'}}}}}}}}}}'}
     }

KNOWN_AMSSYMB_SYMBOLS = [r'\triangle', r'\square', r'\lozenge', r'\bigstar']
KNOWN_AMSMATH_SYMBOLS = [r'\text{', r'\negthickspace', r'\dfrac', r'\tfrac',
                         r'\mathstrut']

KNOWN_TEXTCOMP_SYMBOLS = [r'\textdegree', r'\textdollar', r'\textsterling']

THICKNESS_VALUES = [None, 'thin', 'very thin', 'ultra thin', 'thick',
                    'very thick', 'ultra thick']

ARROW_TIPS = [None, '<->', '<-', '->', '-']

DASHPATTERN_VALUES = ['solid', 'dotted', 'densely dotted', 'loosely dotted',
                      'dashed', 'densely dashed', 'loosely dashed',
                      'dash dot', 'densely dash dot', 'loosely dash dot',
                      'dash dot dot', 'densely dash dot dot',
                      'loosely dash dot dot']

__all__ = ['AttrList', 'OptionsList', 'Command', 'DocumentClass',
           'UsePackage', 'UseTikzLibrary']



# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from mathmakerlib.calculus.number import Number
from mathmakerlib.geometry.point import Point
from mathmakerlib.geometry.bipoint import Bipoint
from mathmakerlib.geometry.angle import AngleDecoration
from . import Quadrilateral


class Rectangle(Quadrilateral):
    """Rectangles."""

    def __init__(self, *points, start_vertex=None, name=None,
                 width=Number(1), length=Number(2),
                 mark_right_angles=True,
                 draw_vertices=False, label_vertices=True,
                 thickness='thick', color=None, rotation_angle=0,
                 winding=None, sloped_sides_labels=True):
        r"""
        Initialize Rectangle.

        This will be done either using the points (if provided) of the points
        parameter or using the keyword arguments start_vertex, width...

        :param points: a list of 4 points that will be the vertices of the
        Rectangle. It is possible to provide no Point at all.
        :type points: a list of Points
        :param start_vertex: the vertex to start to draw the Rectangle
        (default (0; 0))
        :type start_vertex: Point
        :param name: the name of the Rectangle, like ABCD.
        Can be either None (the names will be automatically created), or a
        string of the letters to use to name the vertices. Only single letters
        are supported as Points' names so far (at Polygon's creation).
        See issue #3.
        :type name: None or str
        :param width: the width that will be used to calculate the coordinates
        of the vertices used to build the Rectangle
        :type width: a number
        :param length: the length that will be used to calculate the
        coordinates of the vertices used to build the Rectangle
        :type length: a number
        :param mark_right_angles: if True (default), all four angles will be
        automatically marked as right angles.
        :type mark_right_angles: bool
        :param draw_vertices: whether to actually draw, or not, the vertices
        :type draw_vertices: bool
        :param label_vertices: whether to label, or not, the vertices
        :type label_vertices: bool
        :param thickness: the thickness of the Polygon's sides
        :type thickness: str
        :param color: the color of the Polygon's sides
        :type color: str
        :param rotate: the angle of rotation around isobarycenter
        :type rotate: int
        """
        if points:
            (v0, v1, v2, v3) = points
            self._length = Bipoint(v0, v1).length
            self._width = Bipoint(v1, v2).length
        else:  # 2D Rectangles only (so far)
            if start_vertex is None:
                start_vertex = Point(0, 0)
            # Accepted type for width and length is number, will be checked at
            # vertices' instanciations.
            self._width = width
            self._length = length
            v0 = start_vertex
            v1 = Point(length + start_vertex.x, start_vertex.y)
            v2 = Point(length + start_vertex.x, width + start_vertex.y)
            v3 = Point(start_vertex.x, width + start_vertex.y)
        Quadrilateral.__init__(self, v0, v1, v2, v3, name=name,
                               draw_vertices=draw_vertices,
                               label_vertices=label_vertices,
                               thickness=thickness, color=color,
                               rotation_angle=rotation_angle,
                               winding=winding,
                               sloped_sides_labels=sloped_sides_labels)
        self._type = 'Rectangle'
        if mark_right_angles:
            for a in self.angles:
                a.decoration = AngleDecoration(thickness=thickness)
                a.mark_right = True

    @property
    def width(self):
        return self._width

    @property
    def length(self):
        return self._length

    @property
    def area(self):
        return self.width * self.length

    def setup_labels(self, labels=None, linesegments=None, masks=None):
        """
        Convenience method to easily setup Rectangle's length and width labels.

        If masks is None, then by default, only sides[1] (width) and sides[2]
        (length) labels will be shown. The two others will be masked.

        If labels has only two elements, then the first one will be considered
        as the width and the second one as the length. They will be
        appropriately duplicated.

        :param labels: None or the list of the labels
        :type labels: None or list of 2 or 4 elements
        :param linesegments: the list of the LineSegments to label
        (defaults to Polygon's sides)
        :type linesegments: list (of LineSegments)
        :param masks: the list of masks to setup.
        :type masks: None or list of 4 elements
        """
        if linesegments is None:
            linesegments = self.sides
        if len(labels) == 2:
            labels = [labels[1], labels[0], labels[1], labels[0]]
        if masks is None:
            masks = [' ', None, None, ' ']
        super().setup_labels(labels=labels, linesegments=linesegments,
                             masks=masks)

    @property
    def lbl_width(self):
        if isinstance(self.sides[1].label_value, Number):
            if isinstance(self.sides[3].label_value, Number):
                if self.sides[1].label_value == self.sides[3].label_value:
                    return self.sides[1].label_value
                else:
                    raise ValueError('Two different labels have been set '
                                     'for the width: {} and {}.'
                                     .format(repr(self.sides[1].label_value),
                                             repr(self.sides[3].label_value)))
            else:
                return self.sides[1].label_value
        elif isinstance(self.sides[3].label_value, Number):
            return self.sides[3].label_value
        raise ValueError('No width has been set as a Number.')

    @property
    def lbl_length(self):
        if isinstance(self.sides[0].label_value, Number):
            if isinstance(self.sides[2].label_value, Number):
                if self.sides[0].label_value == self.sides[2].label_value:
                    return self.sides[0].label_value
                else:
                    raise ValueError('Two different labels have been set '
                                     'for the length: {} and {}.'
                                     .format(repr(self.sides[0].label_value),
                                             repr(self.sides[2].label_value)))
            else:
                return self.sides[0].label_value
        elif isinstance(self.sides[2].label_value, Number):
            return self.sides[2].label_value
        raise ValueError('No length has been set as a Number.')

    @property
    def lbl_area(self):
        return self.lbl_width * self.lbl_length



# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from copy import deepcopy
from decimal import Decimal

from mathmakerlib import required
from mathmakerlib.core.word import Word
from mathmakerlib.calculus.exponented import Exponented

LENGTH_UNITS = ['km', 'hm', 'dam', 'm', 'dm', 'cm', 'mm', 'µm', 'nm', 'pm']
CAPACITY_UNITS = ['kL', 'hL', 'daL', 'L', 'dL', 'cL', 'mL', 'µL', 'nL', 'pL']
MASS_UNITS = ['t', 'kg', 'hg', 'dag', 'g', 'dg', 'cg', 'mg', 'µg', 'ng', 'pg']
VOLUME_CAPACITY_MATCH = ['m', '', '', 'dm', '', '', 'cm', '', '', 'mm']
COMMON_LENGTH_UNITS = LENGTH_UNITS[:-3]
COMMON_CAPACITY_UNITS = CAPACITY_UNITS[1:-3]
COMMON_MASS_UNITS = MASS_UNITS[:-3]
ANGLE_UNITS = ['\\textdegree']
CURRENCY_UNITS = ['€', r'\officialeuro',
                  '$', r'\textdollar',
                  '£', r'\textsterling']
CURRENCIES_DICT = {'euro': r'\officialeuro',
                   'dollar': r'\textdollar',
                   'sterling': r'\textsterling'}
AVAILABLE_UNITS = LENGTH_UNITS + CAPACITY_UNITS + MASS_UNITS + ANGLE_UNITS\
    + CURRENCY_UNITS
UNIT_KINDS = {'length': COMMON_LENGTH_UNITS,
              'mass': COMMON_MASS_UNITS,
              'capacity': COMMON_CAPACITY_UNITS,
              'currency': CURRENCY_UNITS}
PHYSICAL_QUANTITIES = {'length': LENGTH_UNITS,
                       'area': LENGTH_UNITS,
                       'volume': LENGTH_UNITS,
                       'capacity': CAPACITY_UNITS,
                       'mass': MASS_UNITS,
                       'currency': CURRENCY_UNITS,
                       'angle': ANGLE_UNITS}


def physical_quantity(u):
    """
    Return the physical quantity matching the given unit.

    :param u: the unit
    :type u: str or Unit
    """
    e = 1
    if isinstance(u, Unit):
        e = u.exponent
        u = str(u.content)
    for pq in PHYSICAL_QUANTITIES:
        if u in PHYSICAL_QUANTITIES[pq]:
            if pq == 'length':
                if e == 2:
                    return 'area'
                elif e == 3:
                    return 'volume'
            return pq
    return None


def difference_of_orders_of_magnitude(unit1, unit2):
    """
    Return the required power of 10 to multiply unit1 by, to get unit2.

    :param unit1: the first unit
    :type unit1: str or Unit
    :param unit2: the second unit (to get from unit1)
    :type unit2: str or Unit
    :rtype: decimal.Decimal
    """
    if isinstance(unit1, str):
        unit1 = Unit(unit1)
    if isinstance(unit2, str):
        unit2 = Unit(unit2)
    phq1 = physical_quantity(unit1)
    phq2 = physical_quantity(unit2)
    if phq1 == phq2:
        dimension = unit1.exponent
        if dimension is None:
            dimension = 1
        magnitude_diff = (PHYSICAL_QUANTITIES[phq2].index(unit2.content)
                          - PHYSICAL_QUANTITIES[phq1].index(unit1.content))\
            * dimension
    elif phq1 == 'volume' and phq2 == 'capacity':
        magnitude_diff = PHYSICAL_QUANTITIES[phq2].index(unit2.content)\
            - VOLUME_CAPACITY_MATCH.index(unit1.content)
    elif phq1 == 'capacity' and phq2 == 'volume':
        magnitude_diff = VOLUME_CAPACITY_MATCH.index(unit2.content)\
            - PHYSICAL_QUANTITIES[phq1].index(unit1.content)
    else:
        raise TypeError('Cannot give the difference of orders of magnitude '
                        'between two units that do not belong to the same '
                        'physical quantity ({} and {}).'
                        .format(unit1, unit2))
    return 10 ** (Decimal(str(magnitude_diff)))


class Unit(Exponented):

    def __init__(self, content, exponent=None):
        if isinstance(content, str):
            Exponented.__init__(self, Word(content),
                                exponent=deepcopy(exponent))
        elif isinstance(content, Unit):
            if exponent is None:
                Exponented.__init__(self, deepcopy(content.content),
                                    exponent=deepcopy(content.exponent))
            else:
                Exponented.__init__(self, deepcopy(content.content),
                                    exponent=deepcopy(exponent))
        else:
            raise TypeError('content must be a str or a Unit. Got {} instead.'
                            .format(str(type(content))))

    def __repr__(self):
        return super().__repr__().replace('Exponented', 'Unit')

    def __eq__(self, other):
        if not isinstance(other, Unit):
            return False
        else:
            return (self.content == other.content
                    and self.exponent == other.exponent)

    def __ne__(self, other):
        if not isinstance(other, Unit):
            return True
        else:
            return not (self.content == other.content
                        and self.exponent == other.exponent)

    def imprint(self, start_expr=True, variant='latex', standalone=True):
        if standalone and variant == 'latex':
            required.package['siunitx'] = True
            if self.content in ['€', r'\officialeuro']:
                required.package['eurosym'] = True
            return r'\si{' + super().imprint(start_expr=start_expr) + '}'
        else:
            return super().imprint(start_expr=start_expr, variant=variant)



# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from mathmakerlib import config
from mathmakerlib.calculus.number import Number
from mathmakerlib.geometry.point import Point
from mathmakerlib.geometry.angle import AngleDecoration
from . import Triangle


class RightTriangle(Triangle):
    """Right Triangles."""

    def __init__(self, start_vertex=None, name=None,
                 leg1_length=Number(2), leg2_length=Number(1),
                 mark_right_angle=True,
                 draw_vertices=False, label_vertices=True,
                 thickness='thick', color=None, rotation_angle=0,
                 winding=None, sloped_sides_labels=True):
        r"""
        Initialize Right Triangle

        :param start_vertex: the vertex to start to draw the Right Triangle
        (default (0; 0))
        :type start_vertex: Point
        :param name: the name of the Triangle, like ABC.
        Can be either None (the names will be automatically created), or a
        string of the letters to use to name the vertices. Only single letters
        are supported as Points' names so far (at Polygon's creation).
        See issue #3.
        :type name: None or str
        :param leg1_length: the leg1's length that will be used to calculate
        the coordinates of the vertices used to build the RightTriangle
        :type leg1_length: a number
        :param leg2_length: the leg2's length that will be used to calculate
        the coordinates of the vertices used to build the RightTriangle
        :type leg2_length: a number
        :param mark_right_angle: if True (default), the right angle will be
        automatically marked as right angle.
        :type mark_right_angle: bool
        :param draw_vertices: whether to actually draw, or not, the vertices
        :type draw_vertices: bool
        :param label_vertices: whether to label, or not, the vertices
        :type label_vertices: bool
        :param thickness: the thickness of the Triangle's sides
        :type thickness: str
        :param color: the color of the Triangle's sides
        :type color: str
        :param rotate: the angle of rotation around isobarycenter
        :type rotate: int
        """
        if start_vertex is None:
            start_vertex = Point(0, 0)
        # Accepted type for leg1's and leg2's lengths is number, will be
        # checked at vertices' instanciations.
        v1 = Point(leg1_length + start_vertex.x, start_vertex.y)
        v2 = Point(leg1_length + start_vertex.x, leg2_length + start_vertex.y)
        if (winding == 'clockwise'
            or (winding is None
                and config.polygons.DEFAULT_WINDING == 'clockwise')):
            start_vertex, v2 = v2, start_vertex
        Triangle.__init__(self, start_vertex, v1, v2, name=name,
                          draw_vertices=draw_vertices,
                          label_vertices=label_vertices,
                          thickness=thickness, color=color,
                          rotation_angle=rotation_angle,
                          winding=winding,
                          sloped_sides_labels=sloped_sides_labels)
        self._type = 'RightTriangle'
        if mark_right_angle:
            self.right_angle.decoration = AngleDecoration(thickness=thickness)
            self.right_angle.mark_right = True

    @property
    def hypotenuse(self):
        return self._sides[2]

    @property
    def right_angle(self):
        return self.angles[1]



# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from abc import ABCMeta


def shoelace_formula(*points):
    x = [v.x for v in points]
    y = [v.y for v in points]
    x.append(x[0])
    y.append(y[0])
    x.append(x[1])
    y.append(y[1])
    return sum(x[i] * (y[i + 1] - y[i - 1])
               for i in range(1, len(points) + 1)) / 2


def check_winding(value):
    if value not in ['clockwise', 'anticlockwise']:
        raise ValueError('Expect \'clockwise\' or \'anticlockwise\'. '
                         'Found \'{}\' instead.'.format(value))


class Oriented(object, metaclass=ABCMeta):
    @property
    def winding(self):
        """Tells whether the Oriented object is clockwise or anticlockwise."""
        return self._winding

    @winding.setter
    def winding(self, value):
        """The winding can be set only once (at object's initialization)."""
        if not hasattr(self, '_winding'):
            check_winding(value)
            setattr(self, '_winding', value)
        else:
            raise AttributeError('Cannot reset the winding of a {}.'
                                 .format(type(self).__name__))



# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from copy import deepcopy

from mathmakerlib import config
from mathmakerlib.calculus import Unit
from mathmakerlib.calculus.tools import is_integer
from mathmakerlib.core.printable import Printable

DEFAULT_CLOCKTIME_CONTEXT = {'h_padding': True, 'min_padding': True,
                             's_padding': True, 'show_0s': True, 'sep': ':',
                             'si_show_0h': True, 'si_show_0min': True,
                             'si_show_0s': True, 'si_only_central': False,
                             'si_zero_as': 's'}


def check_clocktime_context(value):
    if not isinstance(value, dict):
        raise TypeError('context must be a dict. Found {} instead.'
                        .format(type(value)))
    for key in value:
        if key not in DEFAULT_CLOCKTIME_CONTEXT:
            raise KeyError('keys of context must belong to {}; found '
                           '{} instead.'
                           .format(set(DEFAULT_CLOCKTIME_CONTEXT.keys()),
                                   repr(key)))


class ClockTime(Printable):
    """hour:minute:second objects ranging from 00:00:00 to 23:59:59"""

    def __new__(cls, hour=0, minute=0, second=0, context=None):
        if isinstance(hour, ClockTime):
            if context is None:
                context = hour.context
            hour, minute, second = hour.hour, hour.minute, hour.second
        if any(not is_integer(value) for value in [hour, minute, second]):
            raise TypeError('hour, minute and second must be <class \'int\'>. '
                            'Found {}, {} and {} instead.'
                            .format(type(hour), type(minute), type(second)))
        if second < 0:
            minute -= abs(second) // 60 + 1
        if second >= 60:
            minute += second // 60
        second = second % 60
        if minute < 0:
            hour -= abs(minute) // 60 + 1
        if minute >= 60:
            hour += minute // 60
        minute = minute % 60
        hour = hour % 24
        self = super().__new__(cls)
        self._hour, self._minute, self._second = hour, minute, second
        self._context = deepcopy(config.clocktime.CONTEXT)
        if context is not None:
            check_clocktime_context(context)
            self._context.update(context)
        return self

    @property
    def hour(self):
        return self._hour

    @property
    def minute(self):
        return self._minute

    @property
    def second(self):
        return self._second

    @property
    def context(self):
        return self._context

    def __repr__(self):
        return f'ClockTime({self.hour:02}:{self.minute:02}:{self.second:02})'

    def __str__(self):
        return f'{self.hour:02}:{self.minute:02}:{self.second:02}'

    def __eq__(self, other):
        if isinstance(other, ClockTime):
            return (self.hour == other.hour and self.minute == other.minute
                    and self.second == other.second)
        else:
            return False

    def __gt__(self, other):
        return (self.hour, self.minute, self.second) \
            > (other.hour, other.minute, other.second)

    def __lt__(self, other):
        return (self.hour, self.minute, self.second) \
            < (other.hour, other.minute, other.second)

    def __ge__(self, other):
        return (self.hour, self.minute, self.second) \
            >= (other.hour, other.minute, other.second)

    def __le__(self, other):
        return (self.hour, self.minute, self.second) \
            <= (other.hour, other.minute, other.second)

    def __hash__(self):
        return hash(repr(self))

    def __add__(self, other):
        if not isinstance(other, ClockTime):
            raise TypeError('Only a ClockTime can be added to a ClockTime. '
                            'Found {} instead.'.format(type(other)))
        return ClockTime(self.hour + other.hour,
                         self.minute + other.minute,
                         self.second + other.second)

    def __sub__(self, other):
        if not isinstance(other, ClockTime):
            raise TypeError('Only a ClockTime can be subtracted from a '
                            'ClockTime. Found {} instead.'.format(type(other)))
        return ClockTime(self.hour - other.hour,
                         self.minute - other.minute,
                         self.second - other.second)

    # DEFAULT_CLOCKTIME_CONTEXT = {'h_padding': True, 'min_padding': True,
    #                              's_padding': True, 'show_0s': True,
    #                              'sep': ':',
    #                              'si_show_0h': True, 'si_show_0min': True,
    #                              'si_show_0s': True,
    #                              'si_only_central': False,
    #                              'si_zero_as': 's'}
    def imprint(self, start_expr=True, variant='latex'):
        output = []
        zero_h = self.hour == 0
        zero_min = self.minute == 0
        zero_s = self.second == 0
        # /!\ padding is ignored for the first unit to be displayed in the
        # case of 'as_si_units'. E.g. 2 hours will be displayed 2 h, not 02 h
        hours_fmt = '{:02}' if self.context['h_padding'] else '{}'
        minutes_fmt = '{:02}' if self.context['min_padding'] else '{}'
        seconds_fmt = '{:02}' if self.context['s_padding'] else '{}'
        hours = hours_fmt.format(self.hour)
        minutes = minutes_fmt.format(self.minute)
        seconds = seconds_fmt.format(self.second)
        displayed_anything = False
        if self.context['sep'] == 'as_si_units':
            if zero_h and zero_min and zero_s:
                output.append('0')
                output.append(Unit(self.context['si_zero_as']).printed)
            if not zero_h or (zero_h and self.context['si_show_0h']):
                hours = '{}'.format(self.hour)
                output.append(hours)
                output.append(Unit('h').printed)
                displayed_anything = True
            if not zero_min or (zero_min and self.context['si_show_0min']):
                if not displayed_anything:
                    minutes = '{}'.format(self.minute)
                output.append(minutes)
                if not(zero_s and self.context['si_only_central']):
                    output.append(Unit('min').printed)
                displayed_anything = True
            if not zero_s or (zero_s and self.context['si_show_0s']):
                if not displayed_anything:
                    seconds = '{}'.format(self.second)
                output.append(seconds)
                if not(zero_h and self.context['si_only_central']):
                    output.append(Unit('s').printed)
                displayed_anything = True
            output = '~'.join(output)
        else:  # For instance, sep == ':'
            output = [hours, minutes]
            if (zero_s and self.context['show_0s']) or not zero_s:
                output.append(seconds)
            output = self.context['sep'].join(output)
        return output



# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from . import Polygon


class Quadrilateral(Polygon):
    """Quadrilaterals."""

    def __init__(self, *vertices, name=None,
                 draw_vertices=False, label_vertices=True,
                 thickness='thick', color=None, rotation_angle=0,
                 winding=None, sloped_sides_labels=True):
        r"""
        Initialize Quadrilateral

        :param vertices: the vertices of the Quadrilateral
        :type vertices: a list of four Points
        :param name: the name of the Quadrilateral, like ABCD.
        Can be either None (the names will be automatically created), or a
        string of the letters to use to name the vertices. Only single letters
        are supported as Points' names so far (at Polygon's creation).
        See issue #3.
        :type name: None or str
        :param draw_vertices: whether to actually draw, or not, the vertices
        :type draw_vertices: bool
        :param label_vertices: whether to label, or not, the vertices
        :type label_vertices: bool
        :param thickness: the thickness of the Quadrilateral's sides
        :type thickness: str
        :param color: the color of the Quadrilateral's sides
        :type color: str
        :param rotate: the angle of rotation around isobarycenter
        :type rotate: int
        """
        if len(vertices) != 4:
            raise ValueError('Four vertices are required to build a '
                             'Quadrilateral. Found {} instead.'
                             .format(len(vertices)))
        Polygon.__init__(self, *vertices, name=name,
                         draw_vertices=draw_vertices,
                         label_vertices=label_vertices,
                         thickness=thickness, color=color,
                         rotation_angle=rotation_angle,
                         winding=winding,
                         sloped_sides_labels=sloped_sides_labels)
        self._type = 'Quadrilateral'

    def setup_labels(self, labels=None, linesegments=None, masks=None):
        """
        Convenience method to easily setup Quadrilateral's sides' labels.

        If masks is None, then by default, all labels will be masked.

        :param labels: None or the list of the labels
        :type labels: None or list of 4 elements
        :param linesegments: the list of the LineSegments to label
        :type linesegments: list (of LineSegments)
        :param masks: the list of masks to setup. If None (default), all masks
        will be set to None.
        :type masks: None or list of 4 elements
        """
        if labels is None:
            labels = [None, None, None, None]
        if masks is None:
            masks = [None, None, None, None]
        if len(labels) != 4:
            raise ValueError('All four labels must be setup. Found {} values '
                             'instead.'.format(len(labels)))
        if len(masks) != 4:
            raise ValueError('All four masks must be setup. Found {} values '
                             'instead.'.format(len(masks)))
        if linesegments is None:
            linesegments = self.sides
        super().setup_labels(labels=labels, linesegments=linesegments,
                             masks=masks)



# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from decimal import Decimal, InvalidOperation
from abc import ABCMeta, abstractmethod

from mathmakerlib import required
from mathmakerlib.LaTeX import DEFAULT_FONT_SIZES
from mathmakerlib.LaTeX import DEFAULT_COLOR_NAMES, XCOLOR_BASE
from mathmakerlib.LaTeX import XCOLOR_DVIPSNAMES, THICKNESS_VALUES, ARROW_TIPS
from mathmakerlib.core.printable import Printable
from mathmakerlib.calculus.tools import is_number
from mathmakerlib.calculus.number import Number


def check_color(value):
    if value is None or value in DEFAULT_COLOR_NAMES or value in XCOLOR_BASE:
        # Base LaTeX colors do not need to be explicitely loaded.
        # As tikz package already loads xcolor base names, it's not
        # necessary to explicitely load them neither.
        pass
    elif value in XCOLOR_DVIPSNAMES:
        required.package['xcolor'] = True
        required.options['xcolor'].add('dvipsnames')
    else:
        raise ValueError('Unknown color name: {}. Only colors from '
                         'xcolor\'s dvipsnames are yet supported.'
                         .format(value))


def check_scale(value, source_name):
    if not is_number(value):
        raise TypeError('The {}\'s scale must be a number.'
                        .format(source_name))


def tikz_approx_position(slope):
    slope %= Number(360)
    # Caution: modulo on negative Decimals does not behave as on ints.
    # So, it's necessary to add 360 in case of a negative result.
    if slope < 0:
        slope += 360
    if (Decimal('337.5') <= slope <= Decimal('360')
        or Decimal('0') <= slope < Decimal('22.5')):
        return 'right'
    elif Decimal('22.5') <= slope < Decimal('67.5'):
        return 'above right'
    elif Decimal('67.5') <= slope < Decimal('112.5'):
        return 'above'
    elif Decimal('112.5') <= slope < Decimal('157.5'):
        return 'above left'
    elif Decimal('157.5') <= slope < Decimal('202.5'):
        return 'left'
    elif Decimal('202.5') <= slope < Decimal('247.5'):
        return 'below left'
    elif Decimal('247.5') <= slope < Decimal('292.5'):
        return 'below'
    elif Decimal('292.5') <= slope < Decimal('337.5'):
        return 'below right'


def tikz_options_list(options_list, source=None):
    """
    Return '[opt1, opt2,...]' or '' (if all options are None).

    :param options_list: a str keyword (available so far: 'draw') or a list of
    strings (or None)
    :type options_list: list
    :param source: used if options_list is a keyword instead of a list
    :type source: if options_list is 'draw', source must be a Drawable
    :rtype: str
    """
    options = []
    if options_list == 'draw':
        if not isinstance(source, Drawable):
            raise TypeError('Expected a Drawable, found {}.'
                            .format(type(source)))
        options_list = source._tikz_draw_options()
    for o in options_list:
        if o is not None:
            options.append(o)
    if options:
        return '[{}]'.format(', '.join(options))
    else:
        return ''


class Labeled(object, metaclass=ABCMeta):

    @property
    def label_value(self):
        return self._label_value

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, value):
        if value in ['', None]:
            self._label = self._label_value = None
        elif isinstance(value, Printable):
            self._label = value.printed
            self._label_value = value
        else:
            self._label = self._label_value = str(value)


class Colored(object, metaclass=ABCMeta):
    @property
    def color(self):
        if not hasattr(self, '_color'):
            return None
        else:
            return self._color

    @color.setter
    def color(self, value):
        check_color(value)
        setattr(self, '_color', value)


class HasRadius(object, metaclass=ABCMeta):
    @property
    def radius(self):
        if not hasattr(self, '_radius'):
            return None
        return self._radius

    @radius.setter
    def radius(self, value):
        from mathmakerlib.calculus.number import Number
        if is_number(value):
            self._radius = Number(value)
        else:
            raise TypeError('Expected a number as radius. Got {} instead.'
                            .format(str(type(value))))


class HasThickness(object, metaclass=ABCMeta):
    @property
    def thickness(self):
        if not hasattr(self, '_thickness'):
            return None
        return self._thickness

    @thickness.setter
    def thickness(self, value):
        if value in THICKNESS_VALUES:
            self._thickness = value
        else:
            raise ValueError('Incorrect thickness value: \'{}\'. '
                             'Available values belong to: {}.'
                             .format(str(value), str(THICKNESS_VALUES)))


class HasArrowTips(object, metaclass=ABCMeta):
    @property
    def arrow_tips(self):
        if not hasattr(self, '_arrow_tips'):
            return None
        return self._arrow_tips

    @arrow_tips.setter
    def arrow_tips(self, value):
        # TODO: use another mechanism to check possible arrow tips
        # (as it is now, only too few arrow tips are allowed)
        if value in ARROW_TIPS:
            self._arrow_tips = value
        else:
            raise ValueError('Incorrect arrow_tips value: \'{}\'. '
                             'Available values belong to: {}.'
                             .format(str(value), str(ARROW_TIPS)))


class Drawable(Colored, Labeled, metaclass=ABCMeta):

    def draw(self):
        """
        Return the LaTeX (tikz) string of the object.

        The drawing must be made based on the object's properties, no extra
        argument is required. First, setup the object (at initialization, or
        modify it later), once it's ready, draw it.
        """
        required.package['tikz'] = True
        body_format = {}
        section_attr_prefix = 'tikzsection_'
        sections = [attr for attr in dir(self)
                    if attr.startswith(section_attr_prefix)]
        for s in sections:
            body_format.update({s[len(section_attr_prefix):] + '_section':
                                getattr(self, s)()})
        fontsizecomment = 'Text font size'
        body_format.update({'fontsize': '% {}\n{}\n'.format(fontsizecomment,
                                                            self.fontsize)
                                        if self.fontsize
                                        else ''})
        return r"""
\begin{{tikzpicture}}{pic_options}{body}\end{{tikzpicture}}
""".format(pic_options=self.tikz_picture_options(),
           body=self.tikz_picture_body().format(**body_format))

    def tikz_picture_options(self):
        # Prepare possible picture's options
        scale_option = ''
        if self.scale != 1:
            scale_option = 'scale={}'.format(self.scale)
        baseline_option = ''
        if self.baseline is not None:
            baseline_option = 'baseline={}'.format(self.baseline)
        pic_options_list = [_
                            for _ in [baseline_option, scale_option]
                            if _ != '']
        pic_options = ', '.join(pic_options_list)
        if pic_options:
            pic_options = '[{}]'.format(pic_options)
        return pic_options

    def tikz_picture_body(self):
        return r"""
{fontsize}{declarations_section}

{drawing_section}
{labeling_section}{boundingbox_section}
"""

    def tikzsection_declarations(self):
        return '''{declaring_comment}
{declarations}'''.format(declaring_comment=self.tikz_declaring_comment(),
                         declarations=self.tikz_declarations())

    def tikzsection_drawing(self):
        drawing_section = ''
        for (i, (c, d)) in enumerate(zip(self.tikz_drawing_comment(),
                                     self.tikz_draw())):
            drawing_section += ('{{drawing_comment{}}}\n{{drawing{}}}\n'
                                .format(i, i))\
                .format(**{'drawing_comment{}'.format(i): c,
                           'drawing{}'.format(i): d})
        return drawing_section

    def tikzsection_labeling(self):
        return '''{labeling_comment}
{labels}'''.format(labeling_comment=self.tikz_labeling_comment(),
                   labels=self.tikz_points_labels())

    def tikzsection_boundingbox(self):
        boundingbox_section = ''
        if self.boundingbox is not None:
            boundingbox_section = '\n\n' \
                + (r'\useasboundingbox ({},{}) rectangle ({},{});'
                   .format(*self.boundingbox))
        return boundingbox_section

    def tikz_declaring_comment(self):
        """
        Default declaring comment, '% Declare Points'.

        :rtype: str
        """
        return '% Declare Points'

    @abstractmethod
    def tikz_declarations(self):
        """
        Return the necessary declarations (e.g. Points declarations).

        :rtype: str
        """

    @abstractmethod
    def _tikz_draw_options(self):
        """
        The list of possible options for draw command.

        :rtype: list
        """

    @abstractmethod
    def tikz_drawing_comment(self):
        """
        Return the comments matching each drawing category.

        :rtype: list
        """

    @abstractmethod
    def tikz_draw(self):
        """
        Return the commands to actually draw the object.

        They should be grouped in categories (the Points, the Segments etc.).
        Caution, this method must return a list (containing one string per
        category). tikz_drawing_comment() must return a list containing as many
        elements as this one.

        :rtype: list
        """

    def tikz_labeling_comment(self):
        """Default labeling comment, '% Label Points'."""
        return '% Label Points'

    @abstractmethod
    def tikz_label(self):
        """Return the command to write the object's label."""

    @abstractmethod
    def tikz_points_labels(self):
        """Return the command to write the object's points' labels."""

    @property
    def drawn(self):
        """self.drawn is same as self.draw() (no arguments)."""
        return self.draw()

    @property
    def scale(self):
        if not hasattr(self, '_scale'):
            return 1
        else:
            return self._scale

    @scale.setter
    def scale(self, value):
        check_scale(value, 'picture')
        setattr(self, '_scale', value)

    @property
    def baseline(self):
        if not hasattr(self, '_baseline'):
            return None
        else:
            return self._baseline

    @baseline.setter
    def baseline(self, value):
        if value is not None:
            setattr(self, '_baseline', str(value))

    @property
    def fontsize(self):
        if not hasattr(self, '_fontsize'):
            return None
        else:
            return self._fontsize

    @fontsize.setter
    def fontsize(self, value):
        if value is not None:
            if value in DEFAULT_FONT_SIZES:
                setattr(self, '_fontsize', value)
            else:
                raise ValueError('TikZ font size must be None '
                                 'or belong to {}. Found {} instead.'
                                 .format(DEFAULT_FONT_SIZES, repr(value)))

    @property
    def boundingbox(self):
        if not hasattr(self, '_boundingbox'):
            return None
        else:
            return self._boundingbox

    @boundingbox.setter
    def boundingbox(self, value):
        if value is not None:
            if not isinstance(value, tuple):
                raise TypeError('Expected a tuple, found a {} instead.'
                                .format(type(value).__name__))
            if len(value) != 4:
                raise ValueError('Expected a tuple of 4 elements, found {} '
                                 'elements instead.'.format(len(value)))
            for v in value:
                try:
                    Number(v)
                except (TypeError, InvalidOperation):
                    raise TypeError('Expected a tuple containing only '
                                    'numbers. Found a {} instead.'
                                    .format(type(v).__name__))
            setattr(self, '_boundingbox', tuple(Number(v) for v in value))
